from __future__ import annotations

from ...base import exceptions as dmde

__all__ = [
    "DJW_Search_Error",
    "DJW_InvalidClass",
    "DomoJupyterWorkspace",
    "DomoJupyterWorkspaces",
]

import datetime as dt
import os
from dataclasses import dataclass, field
from typing import Any

import httpx

from ...auth import (
    DomoAuth,
    DomoFullAuth,
    DomoJupyterAuth,
    DomoJupyterFullAuth,
    DomoJupyterTokenAuth,
    DomoTokenAuth,
)
from ...base.entities import DomoEntity, DomoManager
from ...client.context import RouteContext
from ...routes import jupyter as jupyter_routes
from ...routes.jupyter import JupyterWorkspace_Error
from ...utils import (
    chunk_execution as dmce,
    files as defi,
)
from .. import DomoUser as dmdu
from ..DomoDataset import DomoDataset as dmds
from . import (
    Account as dmac,
)
from .Account import DomoJupyter_Account
from .Content import DomoJupyter_Content
from .DataSource import DomoJupyter_DataSource


class DJW_Search_Error(dmde.ClassError):
    def __init__(
        self,
        cls=None,
        cls_instance=None,
        search_name=None,
        message=None,
        domo_instance=None,
    ):
        super().__init__(
            cls=cls,
            cls_instance=cls_instance,
            message=message or f"unable to find {search_name}",
            entity_id=domo_instance
            or (cls_instance and cls_instance.auth.domo_instance),
        )


class DJW_InvalidClass(dmde.ClassError):
    def __init__(self, cls_instance, message):
        super().__init__(cls_instance=cls_instance, message=message)


@dataclass(eq=False)
class DomoJupyterWorkspace(DomoEntity):
    auth: DomoJupyterAuth = field(repr=False)
    id: str

    name: str
    description: str

    created_dt: dt.datetime
    updated_dt: dt.datetime

    owner: dict
    cpu: str
    memory: int

    last_run_dt: dt.datetime = None
    instances: list[dict] = None

    input_configuration: list[DomoJupyter_DataSource] = field(
        default_factory=lambda: []
    )
    output_configuration: list[DomoJupyter_DataSource] = field(
        default_factory=lambda: []
    )
    account_configuration: list[DomoJupyter_Account] = field(default_factory=lambda: [])
    content: list[DomoJupyter_Content] = field(default_factory=lambda: [])

    collection_configuration: list[dict] = None
    fileshare_configuration: list[dict] = None

    service_location: str = None
    service_prefix: str = None
    raw: dict

    def __post_init__(self):
        self._update_auth_params()

        self.account_configuration.sort()
        self.output_configuration.sort()
        self.input_configuration.sort()

    def __eq__(self, other) -> bool:
        """Check equality based on workspace ID.

        Args:
            other: Object to compare with

        Returns:
            bool: True if both are DomoJupyterWorkspace instances with the same ID
        """
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        return self.id == other.id

    def _update_auth_params(self):
        """extracts service location and prefix from "instance" object"""

        if self.instances:
            res = jupyter_routes.parse_instance_service_location_and_prefix(
                self.instances[0], self.auth.domo_instance
            )
            self.service_location = res["service_location"]
            self.service_prefix = res["service_prefix"]

    @property
    def entity_type(self):
        return "jupyter_workspace"

    @property
    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/ai-services/jupyter"

    def update_auth_to_jupyter_auth(self, jupyter_token: str):
        """
        interacting with jupyter notebook content requires an additional authentication against the jupyter server (jupyter_token)
        currently we must manually scrape the token from the browser
        """

        self._update_auth_params()

        if isinstance(self.auth, DomoJupyterAuth):
            pass

        elif isinstance(self.auth, DomoFullAuth):
            self.auth = DomoJupyterFullAuth.convert_auth(
                auth=self.auth,
                service_location=self.service_location,
                jupyter_token=jupyter_token,
                service_prefix=self.service_prefix,
            )

        elif isinstance(self.auth, DomoTokenAuth):
            self.auth = DomoJupyterTokenAuth.convert_auth(
                auth=self.auth,
                service_location=self.service_location,
                jupyter_token=jupyter_token,
                service_prefix=self.service_prefix,
            )
        return self.auth

    @classmethod
    def from_dict(
        cls,
        obj,
        auth,
        jupyter_token: str = None,
    ):
        dj_workspace = cls(
            auth=auth,
            id=obj["id"],
            name=obj["name"],
            description=obj["description"],
            created_dt=obj["created"],
            updated_dt=obj["updated"],
            last_run_dt=obj.get("lastRun"),
            instances=obj["instances"],
            owner=obj["owner"],
            memory=obj["memory"],
            cpu=obj["cpu"],
            fileshare_configuration=obj["collectionConfiguration"],
            raw=obj,
        )

        if jupyter_token:
            dj_workspace.update_auth_to_jupyter_auth(jupyter_token=jupyter_token)

        return dj_workspace

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "memory": int(self.memory),
            "cpu": self.cpu,
            "inputConfiguration": [
                confg.to_dict() for confg in self.input_configuration or []
            ],
            "outputConfiguration": [
                confg.to_dict() for confg in self.output_configuration or []
            ],
            "accountConfiguration": [
                confg.to_api() for confg in self.account_configuration or []
            ],
            "fileshareConfiguration": self.collection_configuration or [],
        }

    @classmethod
    async def get_by_id(
        cls,
        workspace_id,
        auth: DomoAuth,  # this API does not require the jupyter_token, but activities inside the workspace will require additional authentication
        jupyter_token=None,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        is_get_config_entities: bool = True,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            **context_kwargs,
        )

        res = await jupyter_routes.get_jupyter_workspace_by_id(
            workspace_id=workspace_id,
            auth=auth,
            context=context,
        )

        if return_raw:
            return res

        djw = cls.from_dict(
            auth=auth,
            obj=res.response,
            jupyter_token=jupyter_token,
        )

        if not is_get_config_entities:
            return djw

        # Load configurations asynchronously
        await djw.get_output_configuration(
            context=context,
            is_suppress_errors=is_suppress_errors,
            is_use_default_account_class=is_use_default_account_class,
        )
        await djw.get_input_configuration(
            context=context,
            is_suppress_errors=is_suppress_errors,
        )
        await djw.get_account_configuration(
            context=context,
            is_suppress_errors=is_suppress_errors,
            is_use_default_account_class=is_use_default_account_class,
        )

        return djw

    async def get_entity_by_id(
        self, entity_id: str, context: RouteContext, **context_kwargs
    ) -> DomoJupyterWorkspace:
        context = RouteContext.build_context(
            context=context,
            session=None,
            debug_api=False,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )
        return await self.get_by_id(
            workspace_id=entity_id, auth=self.auth, context=context
        )

    @classmethod
    async def get_current_workspace(
        cls,
        auth: DomoJupyterAuth,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        *,
        context: RouteContext | None = None,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
        )

        workspace_id = os.getenv("DOMO_WORKSPACE_ID")

        if not workspace_id:
            raise DJW_Search_Error(
                cls=cls,
                message=f"workspace id {workspace_id} not found.  This only works in Domo's Jupyter Workspaces environment",
            )

        return await cls.get_by_id(
            workspace_id=workspace_id, auth=auth, context=context
        )

    async def get_account_configuration(
        self: DomoJupyterWorkspace,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        is_suppress_errors: bool = False,
        is_use_default_account_class: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )

        """Load account configuration from raw data."""
        if self.raw.get("accountConfiguration"):
            from ...base.exceptions import DomoError
            from ...routes.account.exceptions import AccountNoMatchError
            from .. import DomoAccount as dmac

            # Fetch accounts first
            async def fetch_account(ac, context):
                try:
                    return await dmac.DomoAccount.get_by_id(
                        auth=self.auth,
                        account_id=ac["account_id"],
                        is_use_default_account_class=is_use_default_account_class,
                        is_suppress_no_config=is_suppress_errors,
                        context=context,
                        debug_api=debug_api,
                    )
                except (AccountNoMatchError, DomoError):
                    return None

            accounts = await dmce.gather_with_concurrency(
                *[
                    fetch_account(ac, context=context)
                    for ac in self.raw["accountConfiguration"]
                ],
                n=10,
            )

            # Create DomoJupyter_Account instances with fetched accounts
            self.account_configuration = [
                DomoJupyter_Account.from_dict(
                    obj=ac,
                    dj_workspace=self,
                    domo_account=account,
                )
                for ac, account in zip(self.raw["accountConfiguration"], accounts)
            ]
        return self.account_configuration

    async def get_input_configuration(
        self: DomoJupyterWorkspace,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        is_suppress_errors: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )

        """Load input configuration from raw data."""
        if self.raw.get("inputConfiguration"):
            import domolibrary2.routes.dataset as dataset_route

            # Fetch datasets first
            async def fetch_dataset(ic, context):
                try:
                    return await dmds.DomoDataset.get_by_id(
                        auth=self.auth,
                        dataset_id=ic["dataSourceId"],
                        is_suppress_no_config=is_suppress_errors,
                        context=context,
                        debug_api=debug_api,
                    )
                except dataset_route.DatasetNotFoundError:
                    return None

            datasets = await dmce.gather_with_concurrency(
                *[
                    fetch_dataset(ic, context=context)
                    for ic in self.raw["inputConfiguration"]
                ],
                n=10,
            )

            # Create DomoJupyter_DataSource instances with fetched datasets
            self.input_configuration = [
                DomoJupyter_DataSource.from_dict(
                    obj=ic,
                    dj_workspace=self,
                    domo_dataset=dataset,
                )
                for ic, dataset in zip(self.raw["inputConfiguration"], datasets)
            ]
        return self.input_configuration

    async def get_output_configuration(
        self: DomoJupyterWorkspace,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        is_suppress_errors: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )

        """Load output configuration from raw data."""
        if self.raw.get("outputConfiguration"):
            import domolibrary2.routes.dataset as dataset_route

            # Fetch datasets first
            async def fetch_dataset(oc, context):
                try:
                    return await dmds.DomoDataset.get_by_id(
                        auth=self.auth,
                        dataset_id=oc["dataSourceId"],
                        is_suppress_no_config=is_suppress_errors,
                        session=session,
                        debug_api=debug_api,
                        context=context,
                    )
                except dataset_route.DatasetNotFoundError:
                    return None

            datasets = await dmce.gather_with_concurrency(
                *[
                    fetch_dataset(oc, context=context)
                    for oc in self.raw["outputConfiguration"]
                ],
                n=10,
            )

            # Create DomoJupyter_DataSource instances with fetched datasets
            self.output_configuration = [
                DomoJupyter_DataSource.from_dict(
                    obj=oc,
                    dj_workspace=self,
                    domo_dataset=dataset,
                )
                for oc, dataset in zip(self.raw["outputConfiguration"], datasets)
            ]
        return self.output_configuration

    def _add_config(self, config, attribute):
        # print(config.alias)
        config_ls = getattr(self, attribute)

        if config in config_ls:
            for i, ex in enumerate(config_ls):
                if ex == config:
                    config_ls[i] = config

        else:
            config_ls.append(config)

        config_ls.sort()

    def add_config_input_datasource(self, dja_datasource: DomoJupyter_DataSource):
        if not isinstance(dja_datasource, DomoJupyter_DataSource):
            raise DJW_InvalidClass(
                message="must pass instance of DomoJupyter_DataSource",
                cls_instance=self,
            )

        return self._add_config(dja_datasource, attribute="input_configuration")

    def add_config_output_datasource(self, dja_datasource: DomoJupyter_DataSource):
        if not isinstance(dja_datasource, DomoJupyter_DataSource):
            raise DJW_InvalidClass(
                message="must pass instance of DomoJupyter_DataSource",
                cls_instance=self,
            )
        return self._add_config(dja_datasource, attribute="output_configuration")

    def add_config_account(self, dja_account: DomoJupyter_Account):
        if not isinstance(dja_account, DomoJupyter_Account):
            raise DJW_InvalidClass(
                message="must pass instance of DomoJupyter_Account", cls_instance=self
            )
        return self._add_config(dja_account, attribute="account_configuration")

    async def get_content(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        is_recursive: bool = True,
        content_path: str = "",
        ignore_folders: list[str] = None,
        included_filetypes: list[str] = None,
        session: httpx.AsyncClient | None = None,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )

        res = await jupyter_routes.get_content(
            auth=self.auth,
            content_path=content_path,
            ignore_folders=ignore_folders,
            included_filetypes=included_filetypes,
            is_recursive=is_recursive,
            return_raw=return_raw,
            context=context,
        )

        if return_raw:
            return res

        self.content = [
            DomoJupyter_Content.from_dict(obj, auth=self.auth) for obj in res.response
        ]
        return self.content

    async def download_workspace_content(
        self,
        base_export_folder=None,
        replace_folder: bool = True,
        ignore_folders: list[str] = None,
        included_filetypes: list[str] = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        *,
        context: RouteContext | None = None,
    ) -> str:
        """Retrieves content from Domo Jupyter Workspace and downloads to a local folder.

        Args:
            base_export_folder: Base folder path for exports
            replace_folder: Whether to replace existing folder
            ignore_folders: List of folder names to exclude from download
            included_filetypes: List of file extensions to include (e.g., ['.ipynb', '.py', '.md'])
            debug_api: Enable API debugging
            session: Optional httpx client session
        """

        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
        )

        base_export_folder = (
            base_export_folder or f"{self.auth.domo_instance}/{self.name}"
        )

        all_content = await self.get_content(
            ignore_folders=ignore_folders,
            included_filetypes=included_filetypes,
            context=context,
        )
        all_content = [
            content for content in all_content if content.file_type != "directory"
        ]
        defi.upsert_folder(base_export_folder, replace_folder=replace_folder)

        return [
            content.export(default_export_folder=base_export_folder)
            for content in all_content
        ]

    def _test_config_duplicates(self, config_name):
        configuration = getattr(self, config_name)

        if len(set([cfg.alias.lower() for cfg in configuration])) == len(configuration):
            return None

        return f"aliases are not unique for {config_name}"

    async def update_config(
        self,
        config: dict = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            **context_kwargs,
        )

        config = config or self.to_dict()
        try:
            return await jupyter_routes.update_jupyter_workspace_config(
                auth=self.auth,
                workspace_id=self.id,
                config=config,
                context=context,
            )

        except dmde.DomoError as e:
            print(self._test_config_duplicates("account_configuration"))
            print(self._test_config_duplicates("output_configuration"))
            print(self._test_config_duplicates("input_configuration"))
            raise e from e

    async def add_account(
        self,
        domo_account: dmac.DomoAccount,
        dja_account: Any = None,
        domo_user: dmdu.DomoUser = None,
        is_update_config: bool = True,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            **context_kwargs,
        )

        if not dja_account and not isinstance(domo_account, dmac.DomoAccount_Default):
            raise DJW_InvalidClass(
                message="must pass domo_account as class DomoAccount", cls_instance=self
            )

        dja_account = dja_account or DomoJupyter_Account(
            alias=domo_account.name, account_id=domo_account.id, dj_workspace=self
        )

        self.add_config_account(dja_account)

        if not is_update_config:
            return self.account_configuration

        retry = 0
        last_error = None

        while retry <= 1:
            try:
                res = await self.update_config(context=context)

                if return_raw:
                    return res

                account_obj = next(
                    obj
                    for obj in res.response["accountConfiguration"]
                    if obj["alias"] == domo_account.name
                )

                return DomoJupyter_Account(
                    account_id=account_obj["account_id"],
                    alias=account_obj["alias"],
                    dj_workspace=self,
                    domo_account=domo_account,
                )

            except JupyterWorkspace_Error as e:
                last_error = e
                share_user_id = (domo_user and domo_user.id) or (
                    await self.auth.who_am_i()
                ).response["id"]

                await domo_account.Access.share(
                    user_id=share_user_id,
                    debug_api=debug_api,
                    session=session,
                    context=context,
                )

                if retry == 1:
                    raise e from e

                retry += 1

        # This should never be reached due to the logic above, but ensures no implicit None return
        raise (
            last_error
            if last_error
            else JupyterWorkspace_Error(
                operation="add_account",
                workspace_id=self.id,
                message="Unexpected error in add_account retry loop",
            )
        )

    async def add_input_dataset(
        self,
        domo_dataset: dmds.DomoDataset,
        domojupyter_ds: DomoJupyter_DataSource = None,
        domo_user: dmdu.DomoUser = None,
        is_update_config: bool = True,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            **context_kwargs,
        )

        """adds a domo_dataset or domojupyter_dataset to workspace and conditionally updates config"""
        if not domojupyter_ds and not isinstance(domo_dataset, dmds.DomoDataset):
            raise DJW_InvalidClass(
                message="must pass domo_dataset as class DomoDataset", cls_instance=self
            )

        if not domojupyter_ds:
            domojupyter_ds = DomoJupyter_DataSource(
                alias=domo_dataset.name, dataset_id=domo_dataset.id, dj_workspace=self
            )

        self.add_config_input_datasource(domojupyter_ds)

        if not is_update_config:
            return self.input_configuration

        retry = 0
        last_error = None

        while retry <= 1:
            try:
                return await self.update_config(context=context)

            except JupyterWorkspace_Error as e:
                last_error = e
                domo_user = domo_user or await dmdu.DomoUser.get_by_id(
                    auth=self.auth,
                    user_id=(await self.auth.who_am_i()).response["id"],
                    debug_api=debug_api,
                    session=session,
                    context=context,
                )

                await domo_dataset.share(
                    member=domo_user,
                    auth=self.auth,
                    debug_api=debug_api,
                    session=session,
                    context=context,
                )

                if retry == 1:
                    raise e from e

                retry += 1

        # This should never be reached due to the logic above, but ensures no implicit None return
        raise (
            last_error
            if last_error
            else JupyterWorkspace_Error(
                operation="add_input_dataset",
                workspace_id=self.id,
                message="Unexpected error in add_input_dataset retry loop",
            )
        )

    async def add_output_dataset(
        self,
        domo_dataset: dmds.DomoDataset,
        domojupyter_ds: DomoJupyter_DataSource = None,
        domo_user: dmdu.DomoUser = None,
        is_update_config: bool = True,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            **context_kwargs,
        )

        if not domojupyter_ds and not isinstance(domo_dataset, dmds.DomoDataset):
            raise DJW_InvalidClass(
                message="must pass domo_dataset as class DomoDataset", cls_instance=self
            )

        domojupyter_ds = domojupyter_ds or DomoJupyter_DataSource(
            alias=domo_dataset.name, dataset_id=domo_dataset.id, dj_workspace=self
        )

        self.add_config_output_datasource(domojupyter_ds)

        if not is_update_config:
            return self.output_configuration

        retry = 0
        last_error = None

        while retry <= 1:
            try:
                return await self.update_config(context=context)

            except JupyterWorkspace_Error as e:
                last_error = e
                domo_user = domo_user or await dmdu.DomoUser.get_by_id(
                    auth=self.auth,
                    user_id=(await self.auth.who_am_i()).response["id"],
                    debug_api=debug_api,
                    session=session,
                    context=context,
                )

                await domo_dataset.share(
                    member=domo_user,
                    auth=self.auth,
                    debug_api=debug_api,
                    session=session,
                    context=context,
                )

                if retry == 1:
                    raise e from e

                retry += 1

        # This should never be reached due to the logic above, but ensures no implicit None return
        raise (
            last_error
            if last_error
            else JupyterWorkspace_Error(
                operation="add_output_dataset",
                workspace_id=self.id,
                message="Unexpected error in add_output_dataset retry loop",
            )
        )


@dataclass
class DomoJupyterWorkspaces(DomoManager):
    auth: DomoAuth
    workspaces: list[DomoJupyterWorkspace] = None

    parent: Any = None

    @classmethod
    def from_parent(cls, parent):
        return cls(parent=parent, auth=parent.auth)

    async def get(
        self,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            **context_kwargs,
        )

        res = await jupyter_routes.get_jupyter_workspaces(
            auth=self.auth,
            context=context,
        )

        if return_raw:
            return res

        self.workspaces = await dmce.gather_with_concurrency(
            *[
                DomoJupyterWorkspace.get_by_id(
                    auth=self.auth,
                    workspace_id=workspace["id"],
                    context=context,
                    is_use_default_account_class=is_use_default_account_class,
                    is_suppress_errors=is_suppress_errors,
                )
                for workspace in res.response
            ],
            n=10,
        )

        return self.workspaces

    async def search_workspace_by_name(
        self,
        workspace_name: str,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            context=context,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            **context_kwargs,
        )

        res = await self.get(
            context=context,
            return_raw=True,
            is_suppress_errors=is_suppress_errors,
        )

        workspace = next(
            (
                workspace
                for workspace in res.response
                if workspace["name"].lower() == workspace_name.lower()
            ),
            None,
        )

        if not workspace:
            raise DJW_Search_Error(cls_instance=self, search_name=workspace_name)

        if return_raw:
            res.response = workspace
            return res

        return await DomoJupyterWorkspace.get_by_id(
            workspace_id=workspace["id"],
            auth=self.auth,
            context=context,
            is_use_default_account_class=is_use_default_account_class,
            is_suppress_errors=is_suppress_errors,
        )



"""Route-level logging shims.

This module bridges legacy imports like ``from .utils.logging import ...``
that existed before the logging helpers were promoted to ``domolibrary2.utils``.
It simply re-exports the canonical implementations so older modules keep
working without touching every import site.
"""

from ...utils.logging import (  # noqa: F401 (re-exported names)
    ColoredLogger,
    DomoEntityExtractor,
    DomoEntityObjectProcessor,
    DomoEntityProcessor,
    DomoEntityResultProcessor,
    LogDecoratorConfig,
    NoOpEntityExtractor,
    ResponseGetDataProcessor,
    get_colored_logger,
    log_call,
    set_domolibrary_logger,
)

__all__ = [
    "ResponseGetDataProcessor",
    "DomoEntityProcessor",
    "DomoEntityObjectProcessor",
    "DomoEntityExtractor",
    "DomoEntityResultProcessor",
    "NoOpEntityExtractor",
    "ColoredLogger",
    "get_colored_logger",
    "set_domolibrary_logger",
    "log_call",
    "LogDecoratorConfig",
]



from __future__ import annotations

from ..auth import DomoAuth
from ..base.exceptions import RouteError
from ..client import (
    get_data as gd,
)
from ..client.context import RouteContext
from ..client.response import ResponseGetData
from ..utils.logging import (
    DomoEntityExtractor,
    DomoEntityResultProcessor,
    LogDecoratorConfig,
    log_call,
)

__all__ = [
    "RoleNotRetrievedError",
    "Role_CRUD_Error",
    "get_roles",
    "get_role_by_id",
    "get_role_grants",
    "get_role_membership",
    "create_role",
    "delete_role",
    "get_default_role",
    "set_default_role",
    "update_role_metadata",
    "set_role_grants",
    "role_membership_add_users",
]


class RoleNotRetrievedError(RouteError):
    def __init__(
        self,
        res: ResponseGetData,
        message=None,
    ):
        super().__init__(res=res, message=message)


# | export
class Role_CRUD_Error(RouteError):
    def __init__(
        self,
        res: ResponseGetData,
        message=None,
    ):
        super().__init__(res=res, message=message)


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_roles(
    auth: DomoAuth,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
    )

    if not res.is_success:
        raise RoleNotRetrievedError(res=res)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_role_by_id(
    auth: DomoAuth,
    role_id: str,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
    )

    if not res.is_success:
        raise RoleNotRetrievedError(
            res=res,
        )

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_role_grants(
    auth: DomoAuth,
    role_id: str,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/authorities"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
    )

    if len(res.response) == 0:
        role_res = await get_roles(auth=auth, context=context)

        domo_role = [role for role in role_res.response if role.get("id") == role_id]

        if not domo_role:
            raise RoleNotRetrievedError(
                res=res,
                message=f"role {role_id} does not exist",
            )

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_role_membership(
    auth: DomoAuth,
    role_id: str,
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/users"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
    )

    if len(res.response.get("users")) == 0:
        role_res = await get_roles(auth=auth, context=context)

        domo_role = next(
            (role for role in role_res.response if role.get("id") == role_id), None
        )

        if not domo_role:
            raise RoleNotRetrievedError(
                res=res,
                message=f"role {role_id} does not exist or cannot be retrieved",
            )

    if return_raw:
        return res

    res.response = res.response.get("users")

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def create_role(
    auth: DomoAuth,
    name: str,
    description: str,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles"

    body = {"name": name, "description": description}

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=body,
        context=context,
    )

    if not res.is_success:
        raise Role_CRUD_Error(res=res)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def delete_role(
    auth: DomoAuth,
    role_id: int,
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="DELETE",
        context=context,
    )

    if return_raw:
        return res

    if res.status == 400 and res.response == "Bad Request":
        print(
            " 😕 weird API issue, but role should have been deleted.  setting is_success = True \n"
        )
        res.is_success = True

    if not res.is_success:
        raise Role_CRUD_Error(
            res=res,
        )

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def get_default_role(
    auth: DomoAuth,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
):
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/customer-states/user.roleid.default"

    params = {"defaultValue": 2, "ignoreCache": True}

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        params=params,
        context=context,
    )

    if not res.is_success:
        raise RoleNotRetrievedError(res=res)

    res.response = res.response.get("value")

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def set_default_role(
    auth: DomoAuth,
    role_id: str,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    # url = f"https://{auth.domo_instance}.domo.com/api/content/v1/customer-states/user.roleid.default"
    # body = {"name": "user.roleid.default", "value": role_id}

    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/settings"
    body = {"defaultRoleId": int(role_id), "allowlistRoleIds": None}

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=body,
        context=context,
    )

    if not res.is_success:
        raise Role_CRUD_Error(res=res)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def update_role_metadata(
    auth: DomoAuth,
    role_id: str,
    role_name: str,
    role_description: str | None = None,
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
):
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}"

    body = {"name": role_name, "description": role_description, "id": role_id}

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=body,
        context=context,
    )

    if return_raw:
        return res

    if res.status == 400 and res.response == "Bad Request":
        print(
            " 😕 weird API issue, but role should have been modified.  setting is_success = True \n"
        )
        res.is_success = True

    if not res.is_success:
        raise Role_CRUD_Error(res=res)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def set_role_grants(
    auth: DomoAuth,
    role_id: str,
    grants: list[str],
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/authorities"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=grants,
        context=context,
    )

    if return_raw:
        return res

    if res.status == 400 and res.response == "Bad Request":
        print(
            " 😕 weird API issue, but role should have been modified.  setting is_success = True \n"
        )
        res.is_success = True

    if not res.is_success:
        raise Role_CRUD_Error(res=res)

    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def role_membership_add_users(
    auth: DomoAuth,
    role_id: str,
    user_ids: list[str],  # list of user ids
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/users"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=user_ids,
        context=context,
    )

    if not res.is_success:
        raise Role_CRUD_Error(res=res)

    return res



"""Default DomoCard implementation"""

from __future__ import annotations

__all__ = ["DomoCard_Default", "CardDatasets", "Card_DownloadSourceCodeError"]

import json
import os
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any

import httpx

from ...auth import DomoAuth
from ...base.entities import DomoEntity_w_Lineage, DomoManager
from ...base.exceptions import ClassError, DomoError
from ...client.context import RouteContext
from ...routes import card as card_routes
from ...utils import (
    chunk_execution as dmce,
    files as dmfi,
)
from ...utils.logging import (
    DomoEntityObjectProcessor,
    LogDecoratorConfig,
    log_call,
)
from ..DomoGroup.core import (
    DomoGroup,
    DomoGroup as dmgr,
)
from ..DomoUser import DomoUser
from ..subentity.lineage import DomoLineage, register_lineage_type


@register_lineage_type("DomoCard_Default", lineage_type="CARD")
@dataclass
class DomoCard_Default(DomoEntity_w_Lineage):
    """Base DomoCard implementation with core functionality"""

    id: str
    auth: DomoAuth = field(repr=False)
    Lineage: DomoLineage | None = field(repr=False, default=None)
    Datasets: CardDatasets | None = field(repr=False, default=None)

    title: str | None = None
    description: str | None = None
    type: str | None = None
    urn: str | None = None
    chart_type: str | None = None
    dataset_id: str | None = None

    datastore_id: str | None = None

    domo_collections: list[Any] = field(default_factory=list)
    domo_source_code: Any = None

    certification: dict | None = None
    owners: list[Any] = field(default_factory=list)

    @property
    def datasets(self) -> list[Any]:  # DomoDataset
        """Legacy property access - prefer using Datasets.get() for async operations"""
        # This property can't be async, so it returns empty list if not already fetched
        # Users should call await card.Datasets.get() to populate datasets
        return []

    @property
    def entity_type(self):
        return "CARD"

    @property
    def name(self) -> str:
        """Get the display name for this card.

        Implements abstract property from DomoEntity_w_Lineage.
        Cards use the 'title' field as their display name.

        Returns:
            Card title, or "Untitled Card {id}" as fallback
        """
        return self.title or f"Untitled Card {self.id}"

    @property
    def entity_name(self) -> str:
        """Get the display name for this card.

        Cards use the 'title' field as their display name.

        Returns:
            Card title, or card ID as fallback
        """
        return str(self.title) if self.title else self.id

    def __post_init__(self):
        super().__post_init__()
        self.enable_federation_support()
        self.Datasets = CardDatasets(auth=self.auth, parent=self)

    @property
    def display_url(self) -> str:
        return f"https://{self.auth.domo_instance}.domo.com/kpis/details/{self.id}"

    @staticmethod
    def _is_federated(obj: dict[str, Any]) -> bool:
        """
        Returns True if the card metadata indicates federation.

        Checks for an explicit 'isFederated' flag, then scans all datasources for federation indicators
        in 'displayType', 'dataType', or 'providerType'.

        See docs/domo_concepts/federated_and_published_entities.md for details.
        """
        # First check explicit flag
        if obj.get("isFederated") is True:
            return True

        # Then check datasources for federation indicators
        datasources = obj.get("datasources", [])
        if not datasources:
            return False

        for ds in datasources:
            display_type = (ds.get("displayType") or "").upper()
            data_type = (ds.get("dataType") or "").upper()
            provider_type = (ds.get("providerType") or "").upper()

            has_federate = any(
                [
                    "FEDERAT" in display_type,
                    "FEDERAT" in data_type,
                    "FEDERAT" in provider_type,
                ]
            )

            if has_federate:
                return True

        return False

    @property
    def is_federated(self) -> bool:
        """Check if this card is federated"""
        return self._is_federated(self.raw)

    @classmethod
    def from_dict(
        cls, auth: DomoAuth, obj: dict[str, Any], owners: list[Any] | None = None
    ):
        """Synchronous factory method for dict → Card construction.

        Build a DomoCard_Default instance from a metadata dictionary and owners list.

        This method does not invoke any other class methods or perform additional API calls.
        All required data must be provided in the arguments.

        Args:
            auth: DomoAuth instance for authentication
            obj: Card metadata dictionary (from API)
            owners: List of owner entities (DomoUser/DomoGroup), optional

        Returns:
            DomoCard_Default instance
        """
        owners = owners or []

        card = cls(
            auth=auth,
            id=obj.get("id"),
            raw=obj,
            title=obj.get("title"),
            description=obj.get("description"),
            type=obj.get("type"),
            urn=obj.get("urn"),
            certification=obj.get("certification"),
            chart_type=obj.get("metadata", {}).get("chartType"),
            dataset_id=(
                obj.get("datasources", [])[0].get("dataSourceId")
                if obj.get("datasources")
                else None
            ),
            owners=owners,
            datastore_id=obj.get("domoapp", {}).get("id"),
        )

        return card

    @staticmethod
    async def get_owners(
        auth: DomoAuth,
        owners: list[dict[str, Any]],
        is_suppress_errors: bool = True,
        *,
        context: RouteContext | None = None,
    ) -> list[Any]:
        """
        Resolve owner entities (DomoUser/DomoGroup) from a list of owner dicts.

        Args:
            auth: DomoAuth instance
            owners: List of owner dicts (each with 'id' and 'type')
            is_suppress_errors: If True, suppress errors and continue; else raise
            context: Optional RouteContext to preserve session across calls

        Returns:
            List of resolved DomoUser or DomoGroup entities

        Processing steps:
            - For each owner dict, dispatch async get_by_id for USER or GROUP
            - Uses gather_with_concurrency for parallel resolution
            - Suppresses errors if requested, logs suppressed errors
        """
        from .. import (
            DomoUser as dmdu,
        )

        print(owners)
        tasks = []
        for ele in owners:
            try:
                if ele["type"] == "USER":
                    tasks.append(
                        dmdu.DomoUser.get_by_id(
                            auth=auth, user_id=ele["id"], context=context
                        )
                    )
                if ele["type"] == "GROUP":
                    tasks.append(
                        dmgr.DomoGroup.get_by_id(
                            group_id=ele["id"], auth=auth, context=context
                        )
                    )

            except DomoError as e:
                if not is_suppress_errors:
                    raise e from e
                else:
                    print(f"Suppressed error getting owner {ele['id']} - {e}")

        return await dmce.gather_with_concurrency(n=60, *tasks)

    @classmethod
    async def _check_if_published(
        cls,
        auth: DomoAuth,
        card_id: str,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
    ) -> bool:
        """Check if a card is part of a publish/subscribe relationship.

        Simplified approach: If the instance has subscriptions and the card is federated,
        it's likely published. Detailed content checking is expensive and error-prone.

        Args:
            auth: Authentication object
            card_id: Card ID to check
            session: Optional HTTP session
            debug_api: Enable debug logging

        Returns:
            True if card is likely published (instance has subscriptions), False otherwise
        """
        from ...classes.DomoEverywhere import DomoEverywhere

        try:
            # Get all subscriptions for this instance
            domo_everywhere = DomoEverywhere(auth=auth)
            await domo_everywhere.get_subscriptions()

            # If this instance has any subscriptions, federated cards are likely published
            # More detailed checking requires publication entity matching which is expensive
            return len(domo_everywhere.subscriptions) > 0

        except (DomoError, httpx.HTTPError, RuntimeError, ValueError):
            # If check fails completely, default to not published
            return False

    @classmethod
    @log_call(
        level_name="entity",
        config=LogDecoratorConfig(result_processor=DomoEntityObjectProcessor()),
    )
    async def get_by_id(
        cls,
        auth: DomoAuth,
        card_id: str,
        optional_parts: str = "certification,datasources,drillPath,owners,properties,domoapp",
        check_if_published: bool = True,
        parent_auth_retrieval_fn: Callable | None = None,
        parent_auth: DomoAuth | None = None,
        max_subscriptions_to_check: int | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        return_raw: bool = False,
        is_suppress_errors: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        """
        Retrieve a DomoCard by ID, including owners and publication/certification status.

        Args:
            auth: DomoAuth instance
            card_id: Card ID to retrieve
            optional_parts: Comma-separated metadata parts to include (owners, certification, etc.)
            check_if_published: If True, checks if card is published (federated + subscriptions)
            parent_auth_retrieval_fn: Callable returning publisher auth when given publisher domain
            parent_auth: Pre-existing publisher auth (alternative to parent_auth_retrieval_fn)
            max_subscriptions_to_check: Optional limit when scanning subscriptions
            debug_api: Enable API debug logging
            session: Optional httpx session
            return_raw: If True, return raw API response
            is_suppress_errors: If True, suppress errors during owner resolution
            context: Optional RouteContext for API call configuration
            **context_kwargs: Additional context parameters

        Returns:
            DomoCard_Default instance (or raw response if return_raw)

        Processing steps:
            1. Fetch card metadata from API (includes owners, certification, datasources, etc.)
            2. If return_raw, return API response directly
            3. Resolve owners to DomoUser/DomoGroup entities via get_owners (runs separate API calls)
            4. If card is federated and check_if_published is True, run additional check for publication status
            5. Build DomoCard_Default instance with all resolved fields
        """
        context = RouteContext.build_context(
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=1,
            **context_kwargs,
        )

        res = await card_routes.get_card_metadata(
            auth=auth,
            card_id=card_id,
            optional_parts=optional_parts,
            context=context,
        )

        if return_raw:
            return res

        owners = await cls.get_owners(
            auth=auth,
            owners=res.response.get("owners", []),
            is_suppress_errors=is_suppress_errors,
            context=context,
        )

        # Check if published (if federated and check enabled)
        # Auto-enable publish check if auth method is provided
        is_published = False
        subscription = None
        has_auth_method = parent_auth_retrieval_fn or parent_auth
        should_check_publish = check_if_published is True or (
            check_if_published is not False and has_auth_method
        )
        if should_check_publish and cls._is_federated(res.response) and has_auth_method:
            # Create a retrieval function from parent_auth if only parent_auth was provided
            effective_retrieval_fn = parent_auth_retrieval_fn
            if not effective_retrieval_fn and parent_auth:
                effective_retrieval_fn = lambda _domain: parent_auth  # noqa: E731

            # Local import to avoid circular dependency
            from ..subentity.federation_context import FederationContext

            probe = FederationContext.from_entity_id(
                auth=auth,
                entity_id=str(card_id),
                entity_type="CARD",
            )
            is_published = await probe.check_if_published(
                retrieve_parent_auth_fn=effective_retrieval_fn,
                entity_type="CARD",
                session=session,
                debug_api=debug_api,
                max_subscriptions_to_check=max_subscriptions_to_check,
                context=context,
            )
            if is_published:
                subscription = probe.subscription

        domo_card = cls.from_dict(
            auth=auth,
            obj=res.response,
            owners=owners,
            is_published=is_published,
            parent_auth_retrieval_fn=parent_auth_retrieval_fn,
            parent_auth=parent_auth,
        )

        if is_published and subscription:
            helper = domo_card.enable_federation_support()
            helper.hydrate_from_existing(
                subscription=subscription,
                parent_auth_retrieval_fn=parent_auth_retrieval_fn,
                parent_auth=parent_auth,
                content_type="CARD",
                entity_id=str(card_id),
            )

        # Auto-trace lineage if parent_auth_retrieval_fn is provided
        if parent_auth_retrieval_fn and domo_card.Lineage:
            await domo_card.Lineage.get(
                parent_auth_retrieval_fn=parent_auth_retrieval_fn,
                parent_auth=parent_auth,
                session=session,
                debug_api=debug_api,
                context=context,
            )

        return domo_card

    @classmethod
    async def get_entity_by_id(
        cls,
        auth: DomoAuth,
        entity_id: str,
        is_suppress_errors: bool = False,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            debug_num_stacks_to_drop=1,
            **context_kwargs,
        )

        return await cls.get_by_id(
            auth=auth,
            card_id=entity_id,
            is_suppress_errors=is_suppress_errors,
            context=context,
        )

    async def share(
        self,
        auth: DomoAuth | None = None,
        domo_users: list[DomoUser] | None = None,
        domo_groups: list[DomoGroup] | None = None,
        message: str | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ):
        from ...routes import datacenter as datacenter_routes

        context = RouteContext.build_context(
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=1,
            **context_kwargs,
        )

        if domo_groups:
            domo_groups = (
                domo_groups if isinstance(domo_groups, list) else [domo_groups]
            )
        if domo_users:
            domo_users = domo_users if isinstance(domo_users, list) else [domo_users]

        res = await datacenter_routes.share_resource(
            auth=auth or self.auth,
            resource_ids=[self.id],
            resource_type=datacenter_routes.ShareResource_Enum.CARD,
            group_ids=[group.id for group in domo_groups] if domo_groups else None,
            user_ids=[user.id for user in domo_users] if domo_users else None,
            message=message,
            context=context,
        )

        return res

    async def get_collections(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        debug_num_stacks_to_drop: int = 2,
        *,
        context: RouteContext | None = None,
    ):
        from .. import DomoAppDb as dmdb

        context = RouteContext.build_context(
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        domo_collections = await dmdb.AppDbCollections.get_collections(
            datastore_id=self.datastore_id,
            auth=self.auth,
            context=context,
            return_raw=return_raw,
        )

        if return_raw:
            return domo_collections

        self.domo_collections = await dmce.gather_with_concurrency(
            *[
                dmdb.AppDbCollection.get_by_id(
                    collection_id=domo_collection.id,
                    auth=self.auth,
                    debug_api=debug_api,
                    context=context,
                )
                for domo_collection in domo_collections
            ],
            n=60,
        )

        return self.domo_collections

    async def get_source_code(
        self, debug_api: bool = False, try_auto_share: bool = False, **context_kwargs
    ):
        context = RouteContext.build_context(
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )
        await self.get_collections(context=context, debug_api=debug_api)

        collection_name = "ddx_app_client_code"
        code_collection = next(
            (
                domo_collection
                for domo_collection in self.domo_collections
                if domo_collection.name == collection_name
            ),
            None,
        )

        if not code_collection:
            raise Card_DownloadSourceCodeError(
                card=deepcopy(self),
                auth=self.auth,
                message=f"collection - {collection_name} not found for {self.title} - {self.id}",
            )

        documents = await code_collection.query_documents(
            debug_api=debug_api,
            try_auto_share=try_auto_share,
            context=context,
        )

        if not documents:
            raise Card_DownloadSourceCodeError(
                card=deepcopy(self),
                auth=self.auth,
                message=f"collection - {collection_name} - {code_collection.id} - unable to retrieve documents for {self.title} - {self.id}",
            )

        self.domo_source_code = documents[0]

        return self.domo_source_code

    async def download_source_code(
        self,
        download_folder="./EXPORT/",
        file_name=None,
        debug_api: bool = False,
        try_auto_share: bool = False,
        **context_kwargs,
    ):
        context = RouteContext.build_context(
            debug_api=debug_api,
            debug_num_stacks_to_drop=2,
            **context_kwargs,
        )
        doc = await self.get_source_code(
            debug_api=debug_api, try_auto_share=try_auto_share, context=context
        )

        if file_name:
            download_path = os.path.join(
                download_folder, dmfi.change_extension(file_name, new_extension=".json")
            )
            dmfi.upsert_folder(download_path)

            with open(download_path, "w+", encoding="utf-8") as f:
                f.write(json.dumps(doc.content))
                return doc

        ddx_type = next(iter(doc.content))

        for key, value in doc.content[ddx_type].items():
            if key == "js":
                file_name = "app.js"
            elif key == "html":
                file_name = "index.html"
            elif key == "css":
                file_name = "styles.css"
            else:
                file_name = f"{key}.txt"

            download_path = os.path.join(
                download_folder, f"{ddx_type}/{self.id}/{file_name}"
            )
            dmfi.upsert_folder(download_path)

            with open(download_path, "w+", encoding="utf-8") as f:
                f.write(value)

        return doc


@dataclass
class CardDatasets(DomoManager):
    """Manager for datasets associated with a DomoCard

    Provides access to all datasets used by a card through its datasources.
    Inherits from DomoManager to follow standard entity manager patterns.
    """

    auth: DomoAuth = field(repr=False)
    parent: DomoCard_Default = field(repr=False, default=None)

    async def get(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        *,
        context: RouteContext | None = None,
        parent_auth_retrieval_fn: Callable[[str], Any] | None = None,
        **context_kwargs,
    ) -> list[Any]:  # Returns list[DomoDataset]
        """Get all datasets associated with this card

        This retrieves datasets from the card's datasources and returns
        DomoDataset instances for each one.

        Args:
            debug_api: Enable API debugging
            session: Optional httpx session for request reuse

        Returns:
            list[DomoDataset]: List of dataset objects associated with the card
        """
        from ..DomoDataset import DomoDataset

        context = RouteContext.build_context(
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=1,
            **context_kwargs,
        )
        # Get datasources from card metadata if not already loaded
        if not self.parent.raw.get("datasources"):
            # Reload card with datasources
            res = await card_routes.get_card_metadata(
                auth=self.auth,
                card_id=self.parent.id,
                optional_parts="datasources",
                context=context,
            )
            self.parent.raw = res.response

        datasources = self.parent.raw.get("datasources", [])

        if not datasources:
            return []

        # Get dataset IDs from datasources
        dataset_ids = [
            ds.get("dataSourceId") for ds in datasources if ds.get("dataSourceId")
        ]

        if not dataset_ids:
            return []

        # Fetch all datasets concurrently
        # Auto-enable publish check and lineage tracing when parent_auth_retrieval_fn is provided
        datasets = await dmce.gather_with_concurrency(
            *[
                DomoDataset.get_by_id(
                    auth=self.auth,
                    dataset_id=dataset_id,
                    check_if_published=bool(parent_auth_retrieval_fn),
                    parent_auth_retrieval_fn=parent_auth_retrieval_fn,
                    context=context,
                )
                for dataset_id in dataset_ids
            ],
            n=60,
        )

        return datasets


class Card_DownloadSourceCodeError(ClassError):
    def __init__(self, card: DomoCard_Default, auth: DomoAuth, message: str):
        super().__init__(cls_instance=card, message=message, auth=auth)



"""Custom analyzers for domolibrary2-specific patterns."""

from __future__ import annotations

from .domo_visitor import DomoASTVisitor, extend_parser_with_domo_patterns
from .entity_analyzer import EntityAnalyzer
from .mcp_analyzer import MCPAnalyzer
from .route_analyzer import RouteAnalyzer
from .test_analyzer import TestAnalyzer

__all__ = [
    "DomoASTVisitor",
    "extend_parser_with_domo_patterns",
    "EntityAnalyzer",
    "MCPAnalyzer",
    "RouteAnalyzer",
    "TestAnalyzer",
]



"""DomoDatasets Manager Class for collection operations on datasets."""

from dataclasses import dataclass, field

import httpx

from ...auth import DomoAuth
from ...base.entities import DomoManager
from ...client.context import RouteContext
from ...routes import dataset as dataset_routes
from .core import DomoDataset


@dataclass
class DomoDatasets(DomoManager):
    """Manager class for searching and retrieving multiple datasets.

    Provides methods for listing and searching datasets in a Domo instance.
    """

    auth: DomoAuth = field(repr=False)
    datasets: list[DomoDataset] = field(default_factory=list)

    @classmethod
    def _to_domo_dataset(cls, dataset_ls: list, auth: DomoAuth) -> list[DomoDataset]:
        """Convert API response list to DomoDataset objects."""
        return [DomoDataset.from_dict(auth=auth, obj=obj) for obj in dataset_ls]

    async def get(
        self,
        search_text: str | None = None,
        maximum: int | None = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient | None = None,
        return_raw: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ) -> list[DomoDataset]:
        """Retrieve datasets from Domo, optionally filtering by search text.

        Args:
            search_text: Optional text to search for in dataset names
            maximum: Maximum number of datasets to return
            debug_api: Enable API debugging
            debug_num_stacks_to_drop: Stack frames to drop for debugging
            session: HTTP client session (optional, will create one with 30s timeout if not provided)
            return_raw: Return raw API response
            context: Optional RouteContext for API call configuration
            **context_kwargs: Additional context parameters

        Returns:
            List of DomoDataset objects matching the criteria
        """
        # Create session with longer timeout if not provided
        # Datacenter search API can be slow for large result sets
        close_session = False
        if session is None:
            timeout = httpx.Timeout(30.0, connect=10.0)
            session = httpx.AsyncClient(timeout=timeout, verify=False)
            close_session = True

        try:
            context = RouteContext.build_context(
                context=context,
                session=session,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                **context_kwargs,
            )

            res = await dataset_routes.search_datasets(
                auth=self.auth,
                search_text=search_text,
                maximum=maximum,
                context=context,
            )

            if return_raw:
                return res

            self.datasets = self._to_domo_dataset(
                dataset_ls=res.response or [], auth=self.auth
            )
            return self.datasets
        finally:
            if close_session:
                await session.aclose()

    async def search(
        self,
        search_text: str,
        maximum: int | None = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient | None = None,
        return_raw: bool = False,
        *,
        context: RouteContext | None = None,
        **context_kwargs,
    ) -> list[DomoDataset]:
        """Search for datasets by name.

        Args:
            search_text: Text to search for in dataset names (wildcards supported)
            maximum: Maximum number of datasets to return
            debug_api: Enable API debugging
            debug_num_stacks_to_drop: Stack frames to drop for debugging
            session: HTTP client session (optional, will create one with 30s timeout if not provided)
            return_raw: Return raw API response
            context: Optional RouteContext for API call configuration
            **context_kwargs: Additional context parameters

        Returns:
            List of DomoDataset objects matching the search criteria
        """
        return await self.get(
            search_text=search_text,
            maximum=maximum,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            return_raw=return_raw,
            context=context,
            **context_kwargs,
        )



# Classes module
# Users can do: from domolibrary2.classes import DomoUser
# Or: import domolibrary2.classes.DomoUser as DomoUser
# Note: Due to circular imports with client and routes modules,
# individual class modules should be imported directly when needed



"""
Dataflow Tools for Domo MCP Server

Provides tools for managing dataflows in Domo including listing,
monitoring execution history, and triggering dataflow runs.
"""

from mcp.server.fastmcp import Context
from mcp.server.session import ServerSession
from pydantic import BaseModel, Field

from domo_mcp.auth_context import DomoContext
from domo_mcp.server import mcp
from domolibrary2.routes import dataflow as dataflow_routes


class DataflowInfo(BaseModel):
    """Structured output for dataflow metadata."""

    id: int = Field(description="Dataflow ID")
    name: str = Field(description="Dataflow name")
    description: str | None = Field(default=None, description="Dataflow description")
    dataflow_type: str | None = Field(default=None, description="Type of dataflow")
    owner_id: str | None = Field(default=None, description="Owner user ID")
    owner_name: str | None = Field(default=None, description="Owner display name")
    enabled: bool | None = Field(
        default=None, description="Whether dataflow is enabled"
    )
    last_execution_status: str | None = Field(
        default=None, description="Status of last execution"
    )


class DataflowList(BaseModel):
    """Structured output for list of dataflows."""

    dataflows: list[DataflowInfo] = Field(description="List of dataflows")
    total_count: int = Field(description="Total number of dataflows returned")


class DataflowExecution(BaseModel):
    """Structured output for dataflow execution details."""

    execution_id: str = Field(description="Execution ID")
    dataflow_id: int = Field(description="Dataflow ID")
    status: str = Field(description="Execution status")
    start_time: str | None = Field(default=None, description="Execution start time")
    end_time: str | None = Field(default=None, description="Execution end time")
    duration_seconds: int | None = Field(
        default=None, description="Execution duration in seconds"
    )
    error_message: str | None = Field(
        default=None, description="Error message if failed"
    )


class ExecutionHistory(BaseModel):
    """Structured output for dataflow execution history."""

    dataflow_id: int = Field(description="Dataflow ID")
    executions: list[DataflowExecution] = Field(description="List of executions")
    total_count: int = Field(description="Total number of executions")


def _parse_dataflow(data: dict) -> DataflowInfo:
    """Parse dataflow data from API response."""
    owner = data.get("owner", {}) or {}
    return DataflowInfo(
        id=data.get("id", 0),
        name=data.get("name", ""),
        description=data.get("description"),
        dataflow_type=data.get("dapDataFlowType") or data.get("type"),
        owner_id=str(owner.get("id", "")) if owner.get("id") else None,
        owner_name=owner.get("name") or owner.get("displayName"),
        enabled=data.get("enabled"),
        last_execution_status=(
            data.get("lastExecution", {}).get("state")
            if data.get("lastExecution")
            else None
        ),
    )


def _parse_execution(data: dict, dataflow_id: int) -> DataflowExecution:
    """Parse execution data from API response."""
    return DataflowExecution(
        execution_id=str(data.get("id", "")),
        dataflow_id=dataflow_id,
        status=data.get("state", ""),
        start_time=data.get("startedAt"),
        end_time=data.get("endedAt"),
        duration_seconds=data.get("runTime"),
        error_message=data.get("errorMessage"),
    )


@mcp.tool()
async def get_dataflows(
    ctx: Context[ServerSession, DomoContext],
    limit: int = Field(
        default=100, description="Maximum number of dataflows to return"
    ),
) -> DataflowList:
    """List all dataflows in the Domo instance.

    Returns a list of all dataflows with their metadata including
    name, type, owner, and last execution status.
    """
    auth = ctx.request_context.lifespan_context.auth
    await ctx.info(f"Fetching dataflows from {auth.domo_instance}")

    try:
        res = await dataflow_routes.get_dataflows(auth=auth)
        dataflows_data = res.response or []

        # Apply limit
        dataflows_data = dataflows_data[:limit]

        dataflows = [_parse_dataflow(df) for df in dataflows_data]

        await ctx.info(f"Found {len(dataflows)} dataflows")
        return DataflowList(dataflows=dataflows, total_count=len(dataflows))

    except dataflow_routes.GET_Dataflow_Error as e:
        await ctx.error(f"Failed to get dataflows: {e}")
        raise


@mcp.tool()
async def get_dataflow_by_id(
    dataflow_id: int,
    ctx: Context[ServerSession, DomoContext],
) -> DataflowInfo:
    """Get details for a specific dataflow.

    Args:
        dataflow_id: The unique identifier of the dataflow
    """
    auth = ctx.request_context.lifespan_context.auth
    await ctx.info(f"Fetching dataflow {dataflow_id}")

    try:
        res = await dataflow_routes.get_dataflow_by_id(
            auth=auth, dataflow_id=dataflow_id
        )
        return _parse_dataflow(res.response)

    except dataflow_routes.GET_Dataflow_Error as e:
        await ctx.error(f"Failed to get dataflow {dataflow_id}: {e}")
        raise


@mcp.tool()
async def get_dataflow_execution_history(
    dataflow_id: int,
    ctx: Context[ServerSession, DomoContext],
    limit: int = Field(
        default=10, description="Maximum number of executions to return"
    ),
) -> ExecutionHistory:
    """Get the execution history for a dataflow.

    Args:
        dataflow_id: The ID of the dataflow
        limit: Maximum number of executions to return (default: 10)
    """
    auth = ctx.request_context.lifespan_context.auth
    await ctx.info(f"Fetching execution history for dataflow {dataflow_id}")

    try:
        res = await dataflow_routes.get_dataflow_execution_history(
            auth=auth, dataflow_id=dataflow_id
        )
        executions_data = res.response or []

        # Apply limit
        executions_data = executions_data[:limit]

        executions = [_parse_execution(ex, dataflow_id) for ex in executions_data]

        await ctx.info(f"Found {len(executions)} executions")
        return ExecutionHistory(
            dataflow_id=dataflow_id,
            executions=executions,
            total_count=len(executions),
        )

    except dataflow_routes.GET_Dataflow_Error as e:
        await ctx.error(f"Failed to get execution history: {e}")
        raise


@mcp.tool()
async def execute_dataflow(
    dataflow_id: int,
    ctx: Context[ServerSession, DomoContext],
) -> str:
    """Trigger execution of a dataflow.

    Args:
        dataflow_id: The ID of the dataflow to execute
    """
    auth = ctx.request_context.lifespan_context.auth
    await ctx.info(f"Triggering execution for dataflow {dataflow_id}")

    try:
        await dataflow_routes.execute_dataflow(auth=auth, dataflow_id=dataflow_id)

        await ctx.info(f"Dataflow {dataflow_id} execution triggered successfully")
        return f"Successfully triggered execution for dataflow {dataflow_id}"

    except dataflow_routes.CRUD_Dataflow_Error as e:
        await ctx.error(f"Failed to execute dataflow: {e}")
        raise



"""Sample: Track card lineage through federated/published entities.

This sample demonstrates how to use the smart federation resolver to:
1. Get a published card from subscriber instance
2. Automatically discover its subscription
3. Get the parent card from publisher instance (automatically handles indirect publication)
4. Get lineage from both cards
5. Generate a visual mermaid diagram showing the federation chain

Environment Variables Required:
    CONFIG_INSTANCE: Config instance for SDK auth
    CONFIG_ACCESS_TOKEN: Config token for SDK auth
    TEST_CARD_ID: Published card ID in subscriber instance
    TEST_INSTANCE: Subscriber instance name

Key Features:
    - check_if_published=True automatically discovers subscriptions
    - is_fetch_content_details=True automatically loads publication content mapping
    - get_publisher_entity() automatically detects and handles:
      * Directly published entities (cards, pages, datasets)
      * Indirectly published cards (published via pages)
      * Multi-page publications
    - Smart caching prevents redundant API calls
    
Current Limitations:
    - to_mermaid_diagram() with include_federation=True:
      * Automatically resolves directly and indirectly published entities
      * Federated dataset → publisher dataset relationships require BOTH datasets in diagram
      * Uses name-based matching for federated dataset relationships
    - App Studio card resolution not yet implemented (GitHub issue TODO)
"""

import asyncio
import sys
from pathlib import Path
from functools import partial
from dotenv import load_dotenv
import os

# Load environment variables
# Load from parent directory first (common config), then local (overrides)
load_dotenv(dotenv_path=Path(__file__).parent.parent / ".env")  # local_work/work/.env
load_dotenv(dotenv_path=Path(__file__).parent / ".env", override=True)  # local_work/work/lineage/.env (overrides)

# Setup imports
from domolibrary2.classes.DomoCard import DomoCard
from domolibrary2.classes.DomoDataset import DomoDataset
from domolibrary2.utils.logging import get_colored_logger
from domolibrary2.integrations.auth_utils import get_auth_from_env
from domolibrary2.integrations.mermaid import MermaidDiagram
from domolibrary2.integrations.graphs import get_converter
from domolibrary2.auth.base import DomoAuth

# Try to import lutils (local utility module)
# This is required for SDK auth and multi-instance setup
# Path: local_work/work/ -> local_work/work/ (where lutils is)
sys.path.insert(0, str(Path(__file__).parent.parent))  # Add local_work/work/ to path
from lutils.auth import get_sdk_auth_from_codeengine

CARD_ID = os.environ['TEST_CARD_ID']
TARGET_INSTANCE = os.environ['TEST_INSTANCE']
# Optional: Set SHOW_FEDERATION_IDS=true in .env to include publication/subscription IDs
SHOW_FEDERATION_IDS = os.environ.get('SHOW_FEDERATION_IDS', 'false').lower() == 'true'

# Setup logging
logger = get_colored_logger(
    env="development", app_name="track_card_lineage", min_level="WARNING"
)


def normalize_instance_name(value: str) -> str:
    """Normalize Domo instance name to standard format.

    Removes '.domo.com' suffix and 'sdk_' prefix if present.

    Args:
        value: Instance name to normalize

    Returns:
        str: Normalized instance name
    """
    normalized = (value or "").strip()
    if normalized.endswith(".domo.com"):
        normalized = normalized[: normalized.index(".domo.com")]
    if normalized.startswith("sdk_"):
        normalized = normalized[len("sdk_") :]
    return normalized or value


# Test data


async def test_federated_card_lineage_tracking():
    """Test tracking lineage through federated/published card hierarchy.

    This test validates:
    - Published card retrieval from subscriber instance
    - Subscription discovery and validation
    - Parent publication retrieval from publisher
    - Parent card retrieval and lineage
    - Complete lineage chain validation

    Returns:
        DomoCard: The published card with validated lineage

    Raises:
        ValueError: If subscription or parent publication cannot be found
        ImportError: If lutils module is not available
    """

    
    # Setup authentication
    config_auth = await get_auth_from_env(
        domo_instance_env_var="CONFIG_INSTANCE",
        domo_token_env_var="CONFIG_ACCESS_TOKEN",
    )

    # Create partial function for getting auth from any instance
    get_auth = partial(
        get_sdk_auth_from_codeengine,
        config_auth=config_auth,
        use_cache=False,
        debug_api=False,
    )

    # Create auth retrieval function for parent instances
    async def parent_auth_retrieval_fn(publisher_domain: str, **kwargs):
        """Retrieve auth for parent publisher instance.

        Args:
            publisher_domain: Domain of the publisher instance
            **kwargs: Additional context parameters (ignored for simplicity)

        Returns:
            DomoAuth: Authentication for the publisher instance
        """
        normalized = normalize_instance_name(publisher_domain)
        return await get_auth(target_instance=normalized)

    # Get subscriber instance auth
    print(f"📦 Getting auth for playstation (subscriber)...")
    subscriber_auth = await get_auth(target_instance=TARGET_INSTANCE)

    print(f"\n{'='*80}")
    print(f"STEP 1: Retrieve Published Card from Subscriber")
    print(f"{'='*80}")
    print(
        f"Subscriber Instance: {subscriber_auth.domo_instance} | Card ID: {CARD_ID}"
    )

    # Step 1: Get the published card from subscriber
    # NOTE: check_if_published=True will follow subscriptions and may return parent card
    card = await DomoCard.get_by_id(
        auth=subscriber_auth,
        card_id=CARD_ID,
        parent_auth_retrieval_fn=parent_auth_retrieval_fn,
        check_if_published=True,
    )

    # Validate card was retrieved
    assert card is not None, "Card should not be None"
    
    # When check_if_published=True, we may get the parent card instead of subscriber card
    # This is expected behavior - the federation system resolves to the source
    if card.id != CARD_ID:
        print(f"\n⚠️  Note: Card ID resolved from {CARD_ID} to {card.id}")
        print(f"   This means check_if_published followed the subscription to parent card")
    
    assert card.is_federated, "Card should be federated"

    print(f"\n✅ Retrieved Card:")
    print(f"   Requested ID: {CARD_ID}")
    print(f"   Actual ID: {card.id}")
    print(f"   Title: {card.title}")
    print(f"   Type: {type(card).__name__}")
    print(f"   Is Federated: {card.is_federated}")
    print(f"   Instance: {card.auth.domo_instance}")
    print(f"   URL: {card.display_url}")

    # Step 2: Get subscriber lineage
    print(f"\n{'='*80}")
    print(f"STEP 2: Retrieve Subscriber Card Lineage")
    print(f"{'='*80}")

    subscriber_lineage = await card.Lineage.get()

    # Validate subscriber lineage
    assert subscriber_lineage is not None, "Subscriber lineage should not be None"
    assert isinstance(subscriber_lineage, list), "Lineage should be a list"

    print(f"\n✅ Subscriber Lineage ({len(subscriber_lineage)} datasets):")
    for i, link in enumerate(subscriber_lineage, 1):
        if link.entity:
            print(f"   {i}. {link.entity.name}")
            print(f"      Type: {type(link.entity).__name__}")
            print(f"      Is Federated: {link.entity.is_federated}")

    # Step 3: Find subscription
    print(f"\n{'='*80}")
    print(f"STEP 3: Find Subscription")
    print(f"{'='*80}")

    # The subscription was already discovered during check_if_published=True
    # It's stored in card.Federation.subscription
    subscription = card.Federation.subscription if card.Federation else None

    # Validate subscription was found
    if not subscription:
        raise ValueError(
            f"❌ Card {CARD_ID} has NO subscription!\n"
            f"   This card should be a published card from another instance.\n"
            f"   Card type: {type(card).__name__}\n"
            f"   Is federated: {card.is_federated}"
        )

    print(f"\n✅ Subscription Found:")
    print(f"   ID: {subscription.id}")
    print(f"   Publication ID: {subscription.publication_id}")
    print(f"   Subscriber Domain: {subscription.subscriber_domain}")
    print(f"   Publisher Domain: {subscription.publisher_domain}")

    # Step 4: Get publisher card from publisher
    print(f"\n{'='*80}")
    print(f"STEP 4: Retrieve Parent Card from Publisher")
    print(f"{'='*80}")

    # Get publisher auth
    publisher_instance = normalize_instance_name(subscription.publisher_domain)
    print(f"Getting auth for publisher ({publisher_instance})...")
    publisher_auth = await get_auth(target_instance=publisher_instance)

    # Get parent publication with auto-fetched content details
    parent_pub = await card.Federation.get_parent_publication(
        parent_auth=publisher_auth,
        is_fetch_content_details=True,  # Automatic content mapping!
    )
    
    # Show publication details
    print(f"\n📋 Publication content:")
    print(f"   ID: {parent_pub.id}")
    print(f"   Content items: {len(parent_pub.content)}")
    for content in parent_pub.content:
        print(f"      - Type: {content.entity_type}, ID: {content.entity_id}")

    # Get publisher card (automatic direct/indirect resolution)
    parent_card = await card.Federation.get_publisher_entity(
        parent_auth=publisher_auth,
    )
    
    if not parent_card:
        raise ValueError(
            f"❌ Could not retrieve parent card!\n"
            f"   Subscriber card ID: {card.id}\n"
            f"   Publication ID: {subscription.publication_id}"
        )

    print(f"\n✅ Parent Card Retrieved:")
    print(f"   Title: {parent_card.title}")
    print(f"   Type: {type(parent_card).__name__}")
    print(f"   ID: {parent_card.id}")
    print(f"   URL: {parent_card.display_url}")
    
    # Remove redundant publication details since we already showed them above

    # Step 5: Generate subscriber lineage diagram (card → subscription → federated datasets)
    print("\n" + "=" * 80)
    print("STEP 5: Generate Subscriber Lineage Diagram")
    print("=" * 80)

    subscriber_diagram = await card.Lineage.to_mermaid_diagram(
        include_federation=True,
        include_pages=True,
        include_entity_ids=True,
        show_federation_ids=SHOW_FEDERATION_IDS,
        direction="TB",
        parent_auth_retrieval_fn=parent_auth_retrieval_fn,
    )

    # Note: The method now automatically pre-loads publication content lineage
    # including Page → Card relationships, so E → G edge should appear

    # Step 6: Generate Mermaid Diagram
    print(f"\n{'='*80}")
    print(f"STEP 6: Generate Complete Mermaid Diagram")
    print(f"{'='*80}")

    import datetime as dt

    # Use the new graph-based converter architecture
    # 1. Get lineage as a LineageGraph (handles all federation relationships)
    #    - Subscriber card and its lineage
    #    - Subscription and publication nodes
    #    - Publisher card and its lineage (if it can find it)
    #    - Federated dataset → publisher dataset relationships (name-based matching)
    #    - Page nodes for indirect publication
    graph = await card.Lineage.get(
        return_graph=True,
        is_recursive=True,
        parent_auth_retrieval_fn=parent_auth_retrieval_fn,
    )
    
    # 2. Convert graph to Mermaid diagram using the MermaidConverter
    converter = get_converter("mermaid")()
    diagram = converter.convert(
        graph,
        direction="TB",
        include_entity_ids=True,
    )
    
    # NOTE: The built-in method may not find the publisher card for indirectly published
    # entities. If the diagram is missing publisher lineage, that's a known limitation
    # of the automatic resolution. The method successfully handles:
    # 1. Direct publications (card/dataset published directly)
    # 2. Federated dataset relationships (matches by name)
    # But may fail for:
    # 3. Indirectly published cards (cards on published pages)

    # Output diagram to console
    mermaid_output = diagram.to_string()
    print(f"\n{mermaid_output}")

    # Export to file with same name as script
    script_name = Path(__file__).stem  # Gets filename without extension
    markdown_file = Path(__file__).parent / f"{script_name}.md"
    converter.export_to_file(
        diagram,
        str(markdown_file),
        footer_text=f"Generated: {dt.datetime.now().isoformat()}",
    )
    print(f"\n💾 Diagram exported to: {markdown_file}")

    # Final validation summary
    print(f"\n{'='*80}")
    print(f"✅ VALIDATION COMPLETE - ALL STEPS PASSED")
    print(f"{'='*80}")
    print(f"  ✅ Step 1: Retrieved published card from subscriber")
    print(f"  ✅ Step 2: Retrieved subscriber lineage ({len(subscriber_lineage)} datasets)")
    print(f"  ✅ Step 3: Found subscription (ID: {subscription.publication_id})")
    print(f"  ✅ Step 4: Retrieved parent card from publisher")
    print(f"  ✅ Step 5: Retrieved parent card lineage ({len(parent_lineage)} datasets)")
    print(f"  ✅ Step 6: Generated combined Mermaid diagram")
    print(f"     - Subscriber lineage: {len(subscriber_lineage)} datasets")
    print(f"     - Publisher lineage: {len(parent_lineage)} datasets")
    print(f"     - Connected via publication: {subscription.publication_id}")
    print(f"{'='*80}")

    return card


async def main():
    """Main entry point for sample.

    Returns:
        DomoCard: The validated published card
    """


    try:
        card = await test_federated_card_lineage_tracking()
        return card
    except Exception as e:
        print(f"\n❌ Test failed: {e}")
        raise


if __name__ == "__main__":
    card = asyncio.run(main())
    print(f"\n✅ Sample complete!")


"""
DomoDataflow Text Actions

Text manipulation and formatting actions in Magic ETL v2.
These correspond to the "Text" category in the Domo ETL UI sidebar.
"""

from __future__ import annotations

from dataclasses import dataclass

from ....utils import DictDot as util_dd
from .base import DomoDataflow_Action_Base, register_action_type

__all__ = [
    "DomoDataflow_Action_ConcatFields",
    "DomoDataflow_Action_ReplaceString",
    "DomoDataflow_Action_SplitColumn",
    "DomoDataflow_Action_TextFormatting",
]


@register_action_type("ConcatFields", category="text")
@dataclass
class DomoDataflow_Action_ConcatFields(DomoDataflow_Action_Base):
    """Concat Fields action - concatenates multiple columns into one.

    Also known as "Combine Columns" in the Magic ETL UI.

    Attributes:
        fields: List of columns to concatenate
        separator: Separator string between values
        target_field_name: Name of the output column
        remove_selected_fields: Whether to remove source columns

    Example:
        >>> concat_action = dataflow.get_action_objects("ConcatFields")[0]
        >>> print(f"Combining {concat_action.source_columns} with '{concat_action.separator}'")
    """

    fields: list[dict] = None
    separator: str = None
    target_field_name: str = None
    remove_selected_fields: bool = False
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.separator = dd.separator or ""
        self.target_field_name = dd.targetFieldName
        self.remove_selected_fields = dd.removeSelectedFields or False
        self.input = dd.input
        self.tables = dd.tables

    @property
    def source_columns(self) -> list[str]:
        """Get list of columns being concatenated."""
        if not self.fields:
            return []
        return [f.get("name") for f in self.fields if f.get("name")]


@register_action_type("ReplaceString", category="text")
@dataclass
class DomoDataflow_Action_ReplaceString(DomoDataflow_Action_Base):
    """Replace String action - replaces strings in columns.

    Also known as "Replace Text" in the Magic ETL UI.

    Attributes:
        fields: List of replacement configurations

    Example:
        >>> replace_action = dataflow.get_action_objects("ReplaceString")[0]
        >>> for field in replace_action.fields:
        ...     print(f"{field.get('name')}: {field.get('find')} -> {field.get('replace')}")
    """

    fields: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.input = dd.input
        self.tables = dd.tables


@register_action_type("SplitColumnAction", category="text")
@dataclass
class DomoDataflow_Action_SplitColumn(DomoDataflow_Action_Base):
    """Split Column action - splits a column into multiple columns.

    Attributes:
        source_column: Column to split
        delimiter: Delimiter to split on
        delimiter_type: Type of delimiter (STRING, REGEX, etc.)
        use_regex: Whether delimiter is a regex
        additions: New columns to create from splits
        combine_extra_splits: Whether to combine extra splits into last column

    Example:
        >>> split_action = dataflow.get_action_objects("SplitColumnAction")[0]
        >>> print(f"Splitting {split_action.source_column} on '{split_action.delimiter}'")
    """

    source_column: str = None
    delimiter: str = None
    delimiter_type: str = None
    use_regex: bool = False
    additions: list[dict] = None
    combine_extra_splits: bool = False
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.source_column = dd.sourceColumn
        self.delimiter = dd.delimiter
        self.delimiter_type = dd.delimiterType
        self.use_regex = dd.useRegex or False
        self.additions = dd.additions or []
        self.combine_extra_splits = dd.combineExtraSplits or False
        self.input = dd.input
        self.tables = dd.tables

    @property
    def output_columns(self) -> list[str]:
        """Get list of output column names."""
        if not self.additions:
            return []
        return [a.get("name") for a in self.additions if a.get("name")]


@register_action_type("TextFormatting", category="text")
@dataclass
class DomoDataflow_Action_TextFormatting(DomoDataflow_Action_Base):
    """Text Formatting action - applies text formatting transformations.

    Attributes:
        fields: List of text formatting configurations

    Example:
        >>> text_action = dataflow.get_action_objects("TextFormatting")[0]
        >>> for field in text_action.fields:
        ...     print(f"{field.get('name')}: {field.get('operation')}")
    """

    fields: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.input = dd.input
        self.tables = dd.tables



from __future__ import annotations

"""Authentication routes and error handling for Domo API access.

This module provides authentication functions and custom exception classes
for various Domo authentication methods including username/password,
developer tokens, and access tokens.
"""

from typing import Any

from ..base.exceptions import AuthError, RouteError
from ..client import response as rgd
from ..client.context import RouteContext
from ..utils.logging import LogDecoratorConfig, ResponseGetDataProcessor, log_call

__all__ = [
    "AuthError",
    "InvalidCredentialsError",
    "AccountLockedError",
    "InvalidAuthTypeError",
    "InvalidInstanceError",
    "NoAccessTokenReturnedError",
    "get_full_auth",
    "get_developer_auth",
    "who_am_i",
    "elevate_user_otp",
]


class InvalidCredentialsError(RouteError):
    """Raised when invalid credentials are provided to the API."""

    def __init__(self, res=None, **kwargs):
        super().__init__(
            res=res,
            message="Invalid credentials provided",
            **kwargs,
        )


class AccountLockedError(RouteError):
    """Raised when the user account is locked."""

    def __init__(self, res=None, **kwargs):
        super().__init__(
            res=res,
            message="User account is locked",
            **kwargs,
        )


class InvalidAuthTypeError(RouteError):
    """Raised when an invalid authentication type is used for an API call."""

    def __init__(
        self,
        res=None,
        required_auth_type: Any | None = None,
        required_auth_type_ls: list[Any] | None = None,
        **kwargs,
    ):
        # Convert class types to strings
        if required_auth_type:
            required_types = [required_auth_type.__name__]
        elif required_auth_type_ls:
            required_types = [auth_type.__name__ for auth_type in required_auth_type_ls]
        else:
            required_types = ["Unknown"]

        # Build message
        auth_list = ", ".join(required_types)
        message = f"This API requires: {auth_list}"

        super().__init__(
            res=res,
            message=message,
            **kwargs,
        )


class InvalidInstanceError(RouteError):
    """Raised when an invalid Domo instance is provided."""

    def __init__(self, res=None, domo_instance: str | None = None, **kwargs):
        message = (
            f"Invalid Domo instance: {domo_instance}"
            if domo_instance
            else "Invalid Domo instance"
        )

        super().__init__(
            res=res,
            message=message,
            **kwargs,
        )


class NoAccessTokenReturnedError(RouteError):
    """Raised when no access token is returned from the authentication API."""

    def __init__(self, res=None, **kwargs):
        super().__init__(
            res=res,
            message="No access token returned from authentication API",
            **kwargs,
        )


@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def get_full_auth(
    domo_instance: str,  # domo_instance.domo.com
    domo_username: str,  # email address
    domo_password: str,
    auth: Any | None = None,
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """Authenticate using username and password to retrieve a full_auth access token.

    This function uses Domo's standard username/password authentication to obtain
    a session token that can be used for subsequent API calls.

    Args:
        domo_instance (str): The Domo instance identifier
        domo_username (str): User's email address
        domo_password (str): User's password
        auth (Any | None): Existing auth object (optional)
        return_raw (bool): Whether to return raw response without processing
        context (RouteContext | None): Route context for request configuration
        **context_kwargs: Additional context parameters (debug_api, session, parent_class, etc.)

    Returns:
        rgd.ResponseGetData: Response containing session token or error information

    Raises:
        InvalidInstanceError: If the Domo instance is invalid
        InvalidCredentialsError: If credentials are invalid or missing session token
        AccountLockedError: If the user account is locked
        NoAccessTokenReturned: If no access token is returned from the API
    """

    from ..client import get_data as gd
    from ..client.context import RouteContext

    domo_instance = domo_instance or (auth.domo_instance if auth else "")

    url = f"https://{domo_instance}.domo.com/api/content/v2/authentication"

    body = {
        "method": "password",
        "emailAddress": domo_username,
        "password": domo_password,
    }

    context = RouteContext.build_context(context, **context_kwargs)

    res = await gd.get_data(
        auth=auth,  # type: ignore  # Auth can be None for authentication endpoints
        method="POST",
        url=url,
        body=body,
        context=context,
        return_raw=return_raw,
    )

    if return_raw:
        # Type assertion for raw return
        return res  # type: ignore

    # Validate response type
    if not isinstance(res, rgd.ResponseGetData):
        raise TypeError(f"Expected ResponseGetData, got {type(res)}")

    # Handle specific error cases
    if res.status == 403 and res.response == "Forbidden":
        raise InvalidInstanceError(res=res, domo_instance=domo_instance)

    if res.is_success and isinstance(res.response, dict):
        reason = res.response.get("reason")

        if reason == "INVALID_CREDENTIALS":
            res.is_success = False
            raise InvalidCredentialsError(res=res)

        if reason == "ACCOUNT_LOCKED":
            res.is_success = False
            raise AccountLockedError(res=res)

        # Check for empty response
        if res.response == {} or res.response == "":
            res.is_success = False
            raise NoAccessTokenReturnedError(res=res)

    # Validate session token presence
    if isinstance(res.response, dict) and not res.response.get("sessionToken"):
        res.is_success = False
        raise InvalidCredentialsError(res=res)

    return res


@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def get_developer_auth(
    domo_client_id: str,
    domo_client_secret: str,
    auth: Any | None = None,
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """Authenticate using OAuth2 client credentials for developer APIs.

    This function is specifically for authenticating against APIs documented
    under developer.domo.com using OAuth2 client credentials flow.

    Args:
        domo_client_id (str): OAuth2 client ID from developer app registration
        domo_client_secret (str): OAuth2 client secret
        auth (Any | None): Existing auth object (optional)
        return_raw (bool): Whether to return raw response without processing
        context (RouteContext | None): Route context for request configuration
        **context_kwargs: Additional context parameters (debug_api, session, parent_class, etc.)

    Returns:
        rgd.ResponseGetData: Response containing access token or error information

    Raises:
        InvalidCredentialsError: If the client credentials are invalid
    """

    from ..client import get_data as gd
    from ..client.context import RouteContext

    url = "https://api.domo.com/oauth/token?grant_type=client_credentials"

    context = RouteContext.build_context(context, **context_kwargs)

    url = "https://api.domo.com/oauth/token?grant_type=client_credentials"

    res = await gd.get_data(
        method="GET",
        url=url,
        auth=auth,  # type: ignore  # Auth can be None for authentication endpoints
        context=context,
        return_raw=return_raw,
    )

    if return_raw:
        # Type assertion for raw return
        return res  # type: ignore

    # Validate response type
    if not isinstance(res, rgd.ResponseGetData):
        raise TypeError(f"Expected ResponseGetData, got {type(res)}")

    # Handle authentication errors
    if res.status == 401 and res.response == "Unauthorized":
        res.is_success = False
        raise InvalidCredentialsError(res=res)

    return res


@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
    color="cyan",
)
async def who_am_i(
    auth: Any,
    return_raw: bool = False,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """Validate authentication against the 'me' API endpoint.

    This function validates the authentication token by calling Domo's user 'me' API.
    This is the same authentication test the Domo Java CLI uses.

    Args:
        auth (Any): Authentication object containing domo_instance and auth tokens
        return_raw (bool): Whether to return raw response without processing
        context (RouteContext | None): Route context for request configuration
        **context_kwargs: Additional context parameters (debug_api, session, parent_class, etc.)

    Returns:
        rgd.ResponseGetData: Response containing user information or error details

    Raises:
        InvalidInstanceError: If the Domo instance is invalid (403 Forbidden)
        InvalidCredentialsError: If the authentication token is invalid
    """

    from ..client import get_data as gd
    from ..client.context import RouteContext

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/users/me"

    context = RouteContext.build_context(context, **context_kwargs)

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
        return_raw=return_raw,
    )

    if not res.is_success:
        # The @log_call decorator will handle error logging automatically
        pass

    if return_raw:
        # Type assertion for raw return
        return res  # type: ignore

    # Validate response type
    if not isinstance(res, rgd.ResponseGetData):
        raise TypeError(f"Expected ResponseGetData, got {type(res)}")

    # Handle specific error cases
    if res.status == 403 and res.response == "Forbidden":
        raise InvalidInstanceError(res=res)

    if res.status == 401 and res.response == "Unauthorized":
        res.is_success = False  # Fix typo: was is_sucess

    if not res.is_success:
        raise InvalidCredentialsError(res=res)

    return res


@log_call(
    level_name="route",
    config=LogDecoratorConfig(result_processor=ResponseGetDataProcessor()),
)
async def elevate_user_otp(
    auth: Any,
    one_time_password: str,
    user_id: str | None = None,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """Elevate authentication using a one-time password (OTP).

    This function is used when multi-factor authentication is enabled and
    an additional OTP verification step is required.

    Args:
        auth (Any): Authentication object containing domo_instance and tokens
        one_time_password (str): The OTP code for authentication elevation
        user_id (str | None): User ID (will be retrieved from auth if not provided)
        context (RouteContext | None): Route context for request configuration
        **context_kwargs: Additional context parameters (debug_api, session, parent_class, etc.)

    Returns:
        rgd.ResponseGetData: Response from the OTP elevation request

    Raises:
        InvalidCredentialsError: If the OTP is invalid or elevation fails
    """
    context = RouteContext.build_context(context=context, **context_kwargs)

    from ..client import get_data as gd

    # Get user_id from auth if not provided
    if not auth.user_id and not user_id:
        await auth.who_am_i()

    user_id = user_id or auth.user_id

    url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/authentication/elevations/{user_id}"

    body = {"timeBasedOneTimePassword": one_time_password}

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        context=context,
    )

    # Validate response type
    if not isinstance(res, rgd.ResponseGetData):
        raise TypeError(f"Expected ResponseGetData, got {type(res)}")

    if not res.is_success:
        raise InvalidCredentialsError(res=res)

    return res



"""Page-related exception classes."""

__all__ = ["DomoPage_GetRecursive", "Page_NoAccess"]

from ...auth import DomoAuth
from ...base import exceptions as dmde


class DomoPage_GetRecursive(dmde.ClassError):
    def __init__(
        self,
        cls,
        entity_id,
        auth: DomoAuth,
        include_recursive_children,
        include_recursive_parents,
    ):
        super().__init__(
            auth=auth,
            cls=cls,
            entity_id=entity_id,
            message=f"can only trace parents OR children recursively but not both. include_recursive_children : {include_recursive_children}, include_recursive_parents: {include_recursive_parents}",
        )


class Page_NoAccess(dmde.ClassError):
    def __init__(self, page_id, page_title, domo_instance, function_name, parent_class):
        super().__init__(
            function_name=function_name,
            parent_class=parent_class,
            domo_instance=domo_instance,
            message=f'authenticated user doesn\'t have access to {page_id} - "{page_title}" contact owners to share access',
        )



"""
Code Graph MCP Server Package

MCP server that exposes codegraph functionality as AI-accessible tools and resources.
"""

from .server import create_server, mcp

__all__ = ["mcp", "create_server"]



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Any

from ._base import PostmanBase
from .postman_auth import PostmanAuth
from .postman_collection_info import PostmanCollectionInfo
from .postman_event import PostmanEvent
from .postman_folder import PostmanFolder
from .postman_request import PostmanRequest
from .postman_variable import PostmanVariable


@dataclass
class PostmanCollection(PostmanBase):
    """Represents a complete Postman collection.

    This is the root class that contains all the information about
    a Postman collection, including its metadata, folders, requests,
    variables, events, and authentication configuration.

    Attributes:
        info (PostmanCollectionInfo): Collection metadata
        item (list[Union[PostmanRequest, PostmanFolder]]): Collection items (requests and folders)
        auth (Optional[PostmanAuth]): Collection-level authentication
        event (Optional[list[PostmanEvent]]): Collection-level events
        variable (Optional[list[PostmanVariable]]): Collection-level variables
        requests (list[PostmanRequest]): Flat list of all requests (computed property)
    """

    info: PostmanCollectionInfo
    item: list[PostmanRequest | PostmanFolder] = field(default_factory=list)
    auth: PostmanAuth | None = None
    event: list[PostmanEvent] | None = None
    variable: list[PostmanVariable] | None = None

    @property
    def requests(self) -> list[PostmanRequest]:
        """Get a flat list of all requests in the collection."""
        requests = []

        def extract_requests(items):
            for item in items:
                if isinstance(item, PostmanRequest):
                    requests.append(item)
                elif isinstance(item, PostmanFolder):
                    extract_requests(item.item)

        extract_requests(self.item)
        return requests

    @classmethod
    def from_dict(cls, obj: dict[str, Any], **kwargs) -> PostmanCollection:
        """Creates a PostmanCollection object from a JSON dictionary.

        This helper function converts a JSON representation of a Postman
        collection into a structured Python object with proper typing.

        Args:
            obj (dict[str, Any]): The JSON data as a Python dictionary
            **kwargs: Additional keyword arguments (unused)

        Returns:
            PostmanCollection: A structured representation of the collection
        """
        info = PostmanCollectionInfo.from_dict(obj["info"])

        # Process items (can be requests or folders)
        items = []
        for item_data in obj.get("item", []):
            if "request" in item_data:
                # It's a request
                items.append(PostmanRequest.from_dict(item_data))
            else:
                # It's a folder
                items.append(PostmanFolder.from_dict(item_data))

        # Handle collection-level auth
        auth = None
        if obj.get("auth"):
            auth = PostmanAuth.from_dict(obj["auth"])

        # Handle collection-level events
        events = None
        if obj.get("event"):
            events = [PostmanEvent.from_dict(e) for e in obj["event"]]

        # Handle collection-level variables
        variables = None
        if obj.get("variable"):
            variables = [PostmanVariable.from_dict(v) for v in obj["variable"]]

        instance = cls(
            info=info, item=items, auth=auth, event=events, variable=variables
        )
        object.__setattr__(instance, "_raw", obj)
        return instance

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary representation for validation."""
        result = {
            "info": self.info.to_dict(),
        }

        if self.item:
            result["item"] = [item.to_dict() for item in self.item]
        if self.auth:
            result["auth"] = self.auth.to_dict()
        if self.event:
            result["event"] = [e.to_dict() for e in self.event]
        if self.variable:
            result["variable"] = [v.to_dict() for v in self.variable]
        # _raw is excluded as per base class requirement
        return result

    @classmethod
    def from_file(cls, file_path: str) -> PostmanCollection:
        """Load a PostmanCollection from a JSON file.

        Args:
            file_path (str): Path to the Postman collection JSON file

        Returns:
            PostmanCollection: A structured representation of the collection

        Raises:
            FileNotFoundError: If the file doesn't exist
            json.JSONDecodeError: If the file isn't valid JSON
        """

        with open(file_path, encoding="utf-8") as f:
            data = json.load(f)

        return cls.from_dict(data)

    def list_all_headers(self) -> dict[str, list[str]]:
        """List all unique headers and their values from this collection.

        Returns:
            dict[str, list[str]]: Dictionary where keys are header names and values are lists of unique values
        """
        headers_dict = {}

        for request in self.requests:
            for header in request.header:
                if header.key not in headers_dict:
                    headers_dict[header.key] = set()
                headers_dict[header.key].add(header.value)

        # Convert sets to lists for better serialization
        return {key: list(values) for key, values in headers_dict.items()}

    def list_all_params(self) -> dict[str, list[str]]:
        """List all unique query parameters and their values from this collection.

        Returns:
            dict[str, list[str]]: Dictionary where keys are parameter names and values are lists of unique values
        """
        params_dict = {}

        for request in self.requests:
            if request.url.query:
                for param in request.url.query:
                    if param.key not in params_dict:
                        params_dict[param.key] = set()
                    params_dict[param.key].add(param.value)

        # Convert sets to lists for better serialization
        return {key: list(values) for key, values in params_dict.items()}



"""
DomoDataflow Utility Actions

General utility actions for data transformation in Magic ETL v2.
These correspond to the "Utility" category in the Domo ETL UI sidebar.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ....utils import DictDot as util_dd
from .base import DomoDataflow_Action_Base, register_action_type

__all__ = [
    "DomoDataflow_Action_SQL",
    "DomoDataflow_Action_SelectValues",
    "DomoDataflow_Action_Constant",
    "DomoDataflow_Action_ExpressionEvaluator",
    "DomoDataflow_Action_Metadata",
    "DomoDataflow_Action_ValueMapper",
    "DomoDataflow_Action_SetValueField",
]


@register_action_type("SQL", category="utility")
@dataclass
class DomoDataflow_Action_SQL(DomoDataflow_Action_Base):
    """SQL transform action - executes SQL statements.

    Attributes:
        statements: List of SQL statements to execute
        sql_dialect: SQL dialect (e.g., "MAGIC")

    Example:
        >>> sql_action = dataflow.get_action_objects("SQL")[0]
        >>> sql_action.sql = "SELECT * FROM `Input` LIMIT 10"
        >>> dataflow = await dataflow.update_action(sql_action.name, sql_action.raw)
    """

    statements: list[str] = None
    sql_dialect: str = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.statements = dd.statements or []
        if dd.settings:
            self.sql_dialect = dd.settings.get("sqlDialect")

    @property
    def sql(self) -> str | None:
        """Get or set the first SQL statement (convenience property).

        Getting returns the first statement or None.
        Setting updates both self.statements and self.raw.
        """
        if self.statements:
            return self.statements[0]
        return None

    @sql.setter
    def sql(self, value: str | list[str]) -> None:
        """Set SQL statement(s), updating both statements and raw."""
        if isinstance(value, str):
            self.statements = [value]
        else:
            self.statements = list(value)

        # Keep raw in sync
        if self.raw is not None:
            self.raw["statements"] = self.statements


@register_action_type("SelectValues", category="utility")
@dataclass
class DomoDataflow_Action_SelectValues(DomoDataflow_Action_Base):
    """Select Values action - selects/renames/reorders columns.

    Also known as "Select Columns" in the Magic ETL UI.

    Attributes:
        fields: List of field configurations (name, rename, include/exclude)

    Example:
        >>> select_action = dataflow.get_action_objects("SelectValues")[0]
        >>> for field in select_action.fields:
        ...     print(f"{field.get('name')} -> {field.get('rename', field.get('name'))}")
    """

    fields: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.input = dd.input
        self.tables = dd.tables

    @property
    def column_names(self) -> list[str]:
        """Get list of selected column names."""
        if not self.fields:
            return []
        return [f.get("name") for f in self.fields if f.get("name")]

    @property
    def renamed_columns(self) -> dict[str, str]:
        """Get mapping of original name -> renamed name for renamed columns."""
        if not self.fields:
            return {}
        return {
            f.get("name"): f.get("rename")
            for f in self.fields
            if f.get("rename") and f.get("name") != f.get("rename")
        }


@register_action_type("Constant", category="utility")
@dataclass
class DomoDataflow_Action_Constant(DomoDataflow_Action_Base):
    """Constant action - adds constant value columns.

    Also known as "Add Constants" in the Magic ETL UI.

    Attributes:
        fields: List of constant field definitions with name, type, and value

    Example:
        >>> constant_action = dataflow.get_action_objects("Constant")[0]
        >>> for field in constant_action.fields:
        ...     print(f"{field.get('name')}: {field.get('value')}")
    """

    fields: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.input = dd.input
        self.tables = dd.tables

    @property
    def constants(self) -> dict[str, Any]:
        """Get mapping of constant name -> value."""
        if not self.fields:
            return {}
        return {f.get("name"): f.get("value") for f in self.fields if f.get("name")}


@register_action_type("ExpressionEvaluator", category="utility")
@dataclass
class DomoDataflow_Action_ExpressionEvaluator(DomoDataflow_Action_Base):
    """Expression Evaluator action - creates calculated columns using formulas.

    Also known as "Add Formula" or "Beast Mode" in the Magic ETL UI.

    Attributes:
        expressions: List of expression definitions with name, expression, and type

    Example:
        >>> formula_action = dataflow.get_action_objects("ExpressionEvaluator")[0]
        >>> for expr in formula_action.expressions:
        ...     print(f"{expr.get('name')}: {expr.get('expression')}")
    """

    expressions: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.expressions = dd.expressions or []
        self.input = dd.input
        self.tables = dd.tables

    @property
    def formulas(self) -> dict[str, str]:
        """Get mapping of formula name -> expression."""
        if not self.expressions:
            return {}
        return {
            e.get("name"): e.get("expression")
            for e in self.expressions
            if e.get("name") and e.get("expression")
        }


@register_action_type("Metadata", category="utility")
@dataclass
class DomoDataflow_Action_Metadata(DomoDataflow_Action_Base):
    """Metadata action - modifies column metadata (rename, type change).

    Also known as "Set Column Type" or "Rename Columns" in the Magic ETL UI.

    Attributes:
        fields: List of field metadata modifications

    Example:
        >>> metadata_action = dataflow.get_action_objects("Metadata")[0]
        >>> for field in metadata_action.fields:
        ...     print(f"{field.get('name')}: {field.get('type')}")
    """

    fields: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.input = dd.input
        self.tables = dd.tables

    @property
    def type_changes(self) -> dict[str, str]:
        """Get mapping of column -> new type."""
        if not self.fields:
            return {}
        return {f.get("name"): f.get("type") for f in self.fields if f.get("type")}


@register_action_type("ValueMapper", category="utility")
@dataclass
class DomoDataflow_Action_ValueMapper(DomoDataflow_Action_Base):
    """Value Mapper action - maps values from one set to another.

    Also known as "Map Values" in the Magic ETL UI.

    Attributes:
        field_to_use: Source column to map from
        target_field: Target column name
        mappings: List of value mappings
        unmapped_behavior: What to do with unmapped values

    Example:
        >>> mapper_action = dataflow.get_action_objects("ValueMapper")[0]
        >>> print(f"Mapping {mapper_action.field_to_use} -> {mapper_action.target_field}")
    """

    field_to_use: str = None
    target_field: str = None
    target_type: str = None
    mappings: list[dict] = None
    unmapped_behavior: str = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.field_to_use = dd.fieldToUse
        self.target_field = dd.targetField
        self.target_type = dd.targetType
        self.mappings = dd.mappings or []
        self.unmapped_behavior = dd.unmappedBehavior
        self.input = dd.input
        self.tables = dd.tables

    @property
    def value_map(self) -> dict[Any, Any]:
        """Get mapping of source value -> target value."""
        if not self.mappings:
            return {}
        return {
            m.get("source"): m.get("target")
            for m in self.mappings
            if m.get("source") is not None and m.get("target") is not None
        }


@register_action_type("SetValueField", category="utility")
@dataclass
class DomoDataflow_Action_SetValueField(DomoDataflow_Action_Base):
    """Set Value Field action - sets or updates field values.

    Attributes:
        fields: List of field value settings

    Example:
        >>> setvalue_action = dataflow.get_action_objects("SetValueField")[0]
        >>> for field in setvalue_action.fields:
        ...     print(f"{field.get('name')}: {field.get('value')}")
    """

    fields: list[dict] = None
    input: str = None
    tables: list[dict] = None

    def _extract_fields(self, dd: util_dd.DictDot) -> None:
        self.fields = dd.fields or []
        self.input = dd.input
        self.tables = dd.tables



from __future__ import annotations

from ...utils.logging import (
    DomoEntityExtractor,
    DomoEntityResultProcessor,
    LogDecoratorConfig,
    get_colored_logger,
    log_call,
)

logger = get_colored_logger()

"""
CodeEngine CRUD Route Functions

This module provides CRUD functions for managing Domo CodeEngine packages including
creation, deployment, update operations, and account sharing.

Functions:
    deploy_codeengine_package: Deploy a specific package version
    create_codeengine_package: Create a new codeengine package
    increment_version: Increment package version number
    upsert_codeengine_package_version: Create or update a package version
    upsert_package: Create or update a package
    share_accounts_with_package: Share Domo accounts with a package

Classes:
    CodeEnginePackageBuilder: Helper class for building packages (placeholder)
"""

from ...auth import DomoAuth
from ...client import (
    get_data as gd,
    response as rgd,
)
from ...client.context import RouteContext
from . import core as codeengine_routes
from .exceptions import (
    CodeEngine_CRUD_Error,
    CodeEngine_GET_Error,
    CodeEngine_InvalidPackageError,
)
from .packages import generate_share_account_package

__all__ = [
    "CodeEnginePackageBuilder",
    "deploy_codeengine_package",
    "create_codeengine_package",
    "increment_version",
    "upsert_codeengine_package_version",
    "upsert_package",
    "share_accounts_with_package",
]


class CodeEnginePackageBuilder:
    """Helper class for building CodeEngine packages."""

    def __init__(self):
        pass


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def deploy_codeengine_package(
    auth: DomoAuth,
    package_id: str,
    version: str,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    Deploy a specific codeengine package version.

    Args:
        auth: Authentication object
        package_id: Package identifier
        version: Package version to deploy
        context: Route context (optional)
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing deployment result

    Raises:
        CodeEngine_CRUD_Error: If deployment fails
    """
    context = RouteContext.build_context(context=context, **context_kwargs)

    await logger.info(f"Deploying CodeEngine package {package_id} version {version}")

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}/versions/{version}/release"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        context=context,
    )

    if return_raw:
        return res

    if not res.is_success:
        await logger.error(
            f"Failed to deploy package {package_id} v{version}: {res.status_code}"
        )
        raise CodeEngine_CRUD_Error(
            operation="deploy", entity_id=f"{package_id}/v{version}", res=res
        )

    await logger.info(f"Successfully deployed package {package_id} v{version}")
    return res


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def create_codeengine_package(
    auth: DomoAuth,
    payload: dict,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    Create a new codeengine package.

    Args:
        auth: Authentication object
        payload: Package data dictionary
        context: Route context (optional)
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing created package data

    Raises:
        CodeEngine_CRUD_Error: If package creation fails
    """
    context = RouteContext.build_context(context=context, **context_kwargs)

    package_name = payload.get("name", "Unknown")
    package_version = payload.get("version", "Unknown")
    await logger.info(
        f"Creating CodeEngine package '{package_name}' version {package_version}"
    )

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=payload,
        context=context,
    )

    if return_raw:
        return res

    if not res.is_success:
        await logger.error(
            f"Failed to create package '{package_name}': {res.status_code}"
        )
        raise CodeEngine_CRUD_Error(operation="create", res=res)

    package_id = res.response.get("id")
    await logger.info(
        f"Successfully created package '{package_name}' with ID {package_id}"
    )
    return res


def increment_version(version: str) -> str:
    """
    Increment the version number.

    Increments the last part of a dot-separated version string.

    Args:
        version: Version string (e.g., "1.0.0")

    Returns:
        Incremented version string (e.g., "1.0.1")
    """
    parts = version.split(".")
    # Increment the last part
    parts[-1] = str(int(parts[-1]) + 1)
    return ".".join(parts)


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def upsert_codeengine_package_version(
    auth: DomoAuth,
    payload: dict,
    version: str | None = None,
    auto_increment_version: bool = True,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    Create or update a codeengine package version.

    If the package version exists and is released, optionally increment the version.
    If the package code is identical to existing, skip the update.

    Args:
        auth: Authentication object
        payload: Package data dictionary
        version: Package version (optional, defaults to version in payload)
        auto_increment_version: Automatically increment version if deployed
        debug_prn: Enable debug printing
        context: Route context (optional)
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package data

    Raises:
        CodeEngine_InvalidPackageError: If package is already deployed and auto_increment is False
        CodeEngine_CRUD_Error: If package creation/update fails
    """
    context = RouteContext.build_context(context=context, **context_kwargs)

    package_id = payload.get("id")
    version = version or payload.get("version")

    try:
        existing_pkg = await codeengine_routes.get_codeengine_package_by_id_and_version(
            auth=auth,
            package_id=package_id,
            version=version,
            params={"parts": "code"},
            context=context,
        )
        if await codeengine_routes.test_package_is_released(
            existing_package=existing_pkg.response,
            package_id=package_id,
            version=version,
            auth=auth,
        ):
            if not auto_increment_version:
                raise CodeEngine_InvalidPackageError(
                    message=f"Package {package_id} v{version} already deployed",
                    auth=auth,
                )

            version = increment_version(version)

        if await codeengine_routes.test_package_is_identical(
            existing_package=existing_pkg.response,
            new_package=payload,
            package_id=package_id,
            version=version,
            auth=auth,
        ):
            await logger.debug(
                f"Package {package_id} v{version} is identical; skipping update."
            )
            return existing_pkg

    except CodeEngine_GET_Error:
        pass  # Not found, continue to create

    return await create_codeengine_package(
        auth=auth,
        payload=payload,
        context=context,
        return_raw=return_raw,
    )


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def upsert_package(
    auth: DomoAuth,
    payload: dict,
    check_different: bool = True,
    create_new_version: bool = False,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    Create or update a codeengine package.

    If the package doesn't exist, create it. If it exists, update the version.

    Args:
        auth: Authentication object
        payload: Package data dictionary
        check_different: If True, compare payload with existing package and skip
            update if identical (default: True)
        create_new_version: If True, force creation of a new version even if
            package exists (default: False)
        context: Route context (optional)
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package data

    Raises:
        CodeEngine_CRUD_Error: If package creation/update fails
    """
    context = RouteContext.build_context(context=context, **context_kwargs)

    package_id = payload.get("id")
    version = payload.get("version")

    if not package_id:
        await logger.info("No Package ID found, creating new package...")

        return await create_codeengine_package(
            auth=auth,
            payload=payload,
            context=context,
            return_raw=return_raw,
        )

    # Try to get existing package
    await logger.debug(f"Checking if package {package_id} exists")
    try:
        existing_pkg = await codeengine_routes.get_codeengine_package_by_id(
            auth=auth,
            package_id=package_id,
            params={"parts": "code"},
            context=context,
        )
        await logger.debug(f"Found existing package {package_id}")
    except CodeEngine_GET_Error:
        # Package doesn't exist, create it
        await logger.info(f"Package {package_id} not found; creating new package")
        return await create_codeengine_package(
            auth=auth,
            payload=payload,
            context=context,
            return_raw=return_raw,
        )

    # If check_different is enabled, compare with existing package
    if check_different and version:
        await logger.debug(f"Comparing package {package_id} v{version} with existing")
        is_identical = await codeengine_routes.test_package_is_identical(
            existing_package=existing_pkg.response,
            new_package=payload,
            package_id=package_id,
            version=version,
            auth=auth,
            context=context,
        )

        if is_identical:
            await logger.info(
                f"Package {package_id} v{version} is identical to existing; skipping update."
            )
            return existing_pkg

        await logger.debug(
            f"Package {package_id} v{version} differs from existing; proceeding with update"
        )

    # If create_new_version is True, increment version before creating
    if create_new_version and version:
        version = increment_version(version)
        payload["version"] = version
        await logger.info(f"Creating new version {version} for package {package_id}")

    return await upsert_codeengine_package_version(
        payload=payload,
        auth=auth,
        auto_increment_version=not create_new_version,  # Don't auto-increment if we already did
        context=context,
        return_raw=return_raw,
    )


@gd.route_function
@log_call(
    level_name="route",
    config=LogDecoratorConfig(
        entity_extractor=DomoEntityExtractor(),
        result_processor=DomoEntityResultProcessor(),
    ),
)
async def share_accounts_with_package(
    auth: DomoAuth,
    package_id: str,
    account_mappings: list[dict[str, str | int]],
    is_set: bool = True,
    version_id: str | None = None,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    dry_run: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    Share Domo accounts with a CodeEngine package.

    Updates the package manifest to include account mappings, allowing the package
    to access specific Domo accounts by alias.

    Args:
        auth: Authentication object
        package_id: Package identifier
        account_mappings: List of account mapping dicts with 'accountId' and 'alias' keys
            Example: [{"accountId": 92, "alias": "sdk_config"}]
        is_set: If True, replace all existing account mappings with new ones.
            If False, add new mappings to existing ones (default: True)
        context: Route context (optional)
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing updated package data

    Raises:
        CodeEngine_CRUD_Error: If account sharing fails

    Example:
        >>> # Replace all account mappings
        >>> accounts = [
        ...     {"accountId": 92, "alias": "sdk_config"},
        ...     {"accountId": 189, "alias": "sdk_d2c"}
        ... ]
        >>> res = await share_accounts_with_package(
        ...     auth=auth,
        ...     package_id="b368d630-7ca5-4b8a-b4ec-f130cf312dc1",
        ...     account_mappings=accounts,
        ...     is_set=True
        ... )
        >>>
        >>> # Add to existing account mappings
        >>> new_accounts = [{"accountId": 366, "alias": "sdk_audit"}]
        >>> res = await share_accounts_with_package(
        ...     auth=auth,
        ...     package_id="b368d630-7ca5-4b8a-b4ec-f130cf312dc1",
        ...     account_mappings=new_accounts,
        ...     is_set=False
        ... )
    """
    context = RouteContext.build_context(context=context, **context_kwargs)

    await logger.info(
        f"Sharing {len(account_mappings)} account mappings with package {package_id}"
    )

    # First, get the current package version
    if not version_id:
        await logger.debug(f"Retrieving current version for package {package_id}")
        res_version = await codeengine_routes.get_current_package_version(
            auth=auth, package_id=package_id, context=context
        )
        version_id = res_version.response

    if not version_id:
        await logger.error(
            f"Could not determine current version for package {package_id}"
        )
        raise CodeEngine_CRUD_Error(
            operation="share_accounts",
            entity_id=package_id,
            message="Could not determine package version",
        )

    # Get the full version data including code
    res_p = await codeengine_routes.get_codeengine_package_by_id(
        auth=auth,
        package_id=package_id,
        # params={"parts": "code,environment,id,language,manifest,name,version"},
        context=context,
    )

    res_v = await codeengine_routes.get_codeengine_package_by_id_and_version(
        auth=auth,
        package_id=package_id,
        version=version_id,
        # params={"parts": "code,environment,id,language,manifest,name,version"},
        context=context,
    )

    if not is_set:
        # Replace all existing mappings with new ones
        await logger.debug("Merging new account mappings with existing mappings")
        existing_mappings = res_v.response["configuration"].get("accountsMapping", [])
        await logger.debug(
            f"Found {len(existing_mappings)} existing account mappings for version {version_id}"
        )

        # Build dict keyed by accountId for fast lookup
        existing_dict = {
            mapping.get("accountId"): mapping for mapping in existing_mappings
        }

        # Update with new mappings (overwrites if accountId exists, updates alias)
        for new_mapping in account_mappings:
            account_id = new_mapping.get("accountId")
            if account_id:
                existing_dict[account_id] = new_mapping

        # Convert back to list
        account_mappings = list(existing_dict.values())
        await logger.debug(f"Final account mappings count: {len(account_mappings)}")

    body = generate_share_account_package(
        package_id=res_p.response["id"],
        code=res_v.response["code"],
        version=res_v.response["version"],
        name=res_p.response["name"],
        environment=res_p.response["environment"],
        language=res_p.response["language"],
        functions=res_v.response["functions"],
        account_mappings=account_mappings,
    )

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages"

    await logger.debug(
        f"Submitting account mappings for package {package_id} version {version_id}"
    )
    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=body,
        context=context,
    )

    if return_raw:
        return res

    if not res.is_success:
        await logger.error(
            f"Failed to share accounts with package {package_id}: {res.status_code}"
        )
        raise CodeEngine_CRUD_Error(
            operation="share_accounts",
            entity_id=package_id,
            res=res,
        )

    await logger.info(
        f"Successfully shared {len(account_mappings)} account mappings with package {package_id}"
    )
    return res



"""Parsing module for Postman collections.

This module provides the main entry point for parsing Postman collections
into structured models.
"""

from __future__ import annotations

import json
from pathlib import Path

from .models import ParsedPostmanCollection, PostmanCollection, PostmanParseError

__all__ = ["parse_postman_collection"]


def parse_postman_collection(
    collection_path: str | Path,
    base_url: str | None = None,
) -> ParsedPostmanCollection:
    """Parse a Postman collection JSON file into a structured format.

    Args:
        collection_path: Path to Postman collection JSON file
        base_url: Optional base URL to use (if not provided, extracted from collection variables)

    Returns:
        ParsedPostmanCollection with extracted metadata

    Raises:
        PostmanParseError: If collection file cannot be read or parsed
        FileNotFoundError: If collection file doesn't exist
    """
    path = Path(collection_path)
    if not path.exists():
        raise FileNotFoundError(f"Collection file not found: {collection_path}")

    try:
        with open(path, encoding="utf-8") as f:
            data = json.load(f)
    except json.JSONDecodeError as e:
        raise PostmanParseError(f"Invalid JSON in collection file: {e}") from e
    except Exception as e:
        raise PostmanParseError(f"Error reading collection file: {e}") from e

    # Parse using PostmanCollection model
    try:
        collection = PostmanCollection.from_dict(data)
    except Exception as e:
        raise PostmanParseError(
            f"Error parsing Postman collection structure: {e}"
        ) from e

    # Extract base URL from variables if not provided
    if not base_url:
        for var in collection.variable or []:
            if var.key == "baseUrl" or var.key == "instance":
                base_url = var.value or ""
                break

    # Create parsed collection
    parsed = ParsedPostmanCollection(
        collection=collection,
        base_url=base_url or "",
    )

    return parsed



from __future__ import annotations

"""
AppDb Exception Classes

This module contains all exception classes used by AppDb route functions.

Exception Classes:
    AppDb_GET_Error: Raised when AppDb retrieval operations fail
    SearchAppDbNotFoundError: Raised when AppDb search returns no results
    AppDb_CRUD_Error: Raised when AppDb create/update/delete operations fail
"""

from ...base.exceptions import RouteError

__all__ = [
    "AppDb_GET_Error",
    "SearchAppDbNotFoundError",
    "AppDb_CRUD_Error",
]


class AppDb_GET_Error(RouteError):
    """Raised when AppDb retrieval operations fail."""

    def __init__(
        self,
        appdb_id: str | None = None,
        message: str | None = None,
        res=None,
        **kwargs,
    ):
        super().__init__(
            message=message or "AppDb retrieval failed",
            entity_id=appdb_id,
            res=res,
            **kwargs,
        )


class SearchAppDbNotFoundError(RouteError):
    """Raised when AppDb search operations return no results."""

    def __init__(
        self,
        search_criteria: str,
        message: str | None = None,
        res=None,
        **kwargs,
    ):
        super().__init__(
            message=message or f"No AppDb items found matching: {search_criteria}",
            res=res,
            **kwargs,
        )


class AppDb_CRUD_Error(RouteError):
    """Raised when AppDb create, update, or delete operations fail."""

    def __init__(
        self,
        operation: str,
        appdb_id: str | None = None,
        message: str | None = None,
        res=None,
        **kwargs,
    ):
        super().__init__(
            message=message or f"AppDb {operation} operation failed",
            entity_id=appdb_id,
            res=res,
            **kwargs,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ._base import PostmanBase


@dataclass
class PostmanQueryParam(PostmanBase):
    """Represents a URL query parameter.

    Attributes:
        key (str): The parameter name
        value (str): The parameter value
        disabled (bool): Whether the parameter is disabled
        description (str): Parameter description
    """

    key: str
    value: str
    disabled: bool = False
    description: str | None = None

    @classmethod
    def from_dict(cls, obj: dict[str, Any], **kwargs) -> PostmanQueryParam:
        """Create a PostmanQueryParam from parameter data.

        Args:
            obj (dict[str, Any]): Dictionary containing parameter key and value
            **kwargs: Additional keyword arguments (unused)

        Returns:
            PostmanQueryParam: A new query parameter instance
        """
        instance = cls(
            key=obj["key"],
            value=obj["value"],
            disabled=obj.get("disabled", False),
            description=obj.get("description"),
        )
        object.__setattr__(instance, "_raw", obj)
        return instance

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary representation."""
        result = {"key": self.key, "value": self.value}
        if self.disabled:
            result["disabled"] = self.disabled
        if self.description:
            result["description"] = self.description
        # _raw is excluded as per base class requirement
        return result



"""
Module exports
"""

from .core import *

__all__ = [
    "create_fileset",
    "create_folder",
    "delete_fileset",
    "download_file",
    "get_fileset",
    "get_fileset_access",
    "get_fileset_stats",
    "getsearch_files",
    "listsearch_filesets",
    "search_files_with_ai",
    "update_fileset",
    "update_fileset_access",
    "update_fileset_owner",
]



"""
Data Conversion Utilities

This module provides comprehensive utilities for converting between different data formats,
types, and structures. Includes datetime conversions, string formatting, data validation,
and dataframe operations.

Functions:
    # Display utilities
    print_md: Display markdown in Jupyter notebooks

    # Datetime conversions
    convert_epoch_millisecond_to_datetime: Convert epoch milliseconds to datetime
    convert_datetime_to_epoch_millisecond: Convert datetime to epoch milliseconds
    convert_string_to_datetime: Parse string to datetime object

    # Code analysis utilities
    convert_python_to_ast_module: Parse Python code to AST module
    extract_ast_functions: Extract function definitions from AST

    # String formatting utilities
    convert_programming_text_to_title_case: Convert code names to title case
    convert_snake_to_pascal: Convert snake_case to camelCase
    convert_str_to_snake_case: Convert strings to snake_case

    # Validation utilities
    is_valid_email: Validate email format
    convert_string_to_bool: Convert string to boolean

    # Data structure utilities
    concat_list_dataframe: Concatenate list of DataFrames
    merge_dict: Deep merge dictionaries

Exception Classes:
    InvalidEmail: Raised when email validation fails
    ConcatDataframeInvalidElementError: Raised when dataframe concat fails

Example:
    >>> # Datetime conversions
    >>> epoch = convert_datetime_to_epoch_millisecond(datetime.now())
    >>> dt = convert_epoch_millisecond_to_datetime(epoch)

    >>> # String formatting
    >>> title = convert_programming_text_to_title_case("get_user_data")
    >>> # Returns: "Get User Data"

    >>> # Email validation
    >>> try:
    ...     is_valid_email("user@example.com")  # Returns True
    ... except InvalidEmail:
    ...     print("Invalid email")
"""

__all__ = [
    "print_md",
    "convert_epoch_millisecond_to_datetime",
    "convert_datetime_to_epoch_millisecond",
    "convert_string_to_datetime",
    "convert_python_to_ast_module",
    "extract_ast_functions",
    "convert_programming_text_to_title_case",
    "convert_snake_to_pascal",
    "convert_str_to_snake_case",
    "InvalidEmailError",
    "is_valid_email",
    "convert_string_to_bool",
    "ConcatDataframeError",
    "concat_list_dataframe",
    "merge_dict",
]


import ast
import datetime as dt
import re

# Optional dependencies with fallbacks
import pandas as pd
from dateutil import parser as date_parser
from IPython.display import display_markdown

# Import custom exceptions
from .exceptions import ConcatDataframeError, InvalidEmailError


## Legacy exception aliases removed for N818 compliance
def print_md(md_str: str) -> None:
    """
    Display markdown string in Jupyter notebook environment.

    Args:
        md_str (str): Markdown string to display

    Raises:
        ImportError: If IPython is not available

    Example:
        >>> print_md("# Header\\n**Bold text**")

    Note:
        This function only works in Jupyter notebook environments where
        IPython.display is available.
    """
    display_markdown(md_str, raw=True)


def convert_epoch_millisecond_to_datetime(
    epoch: int | None,
) -> dt.datetime | None:
    """
    Convert Epoch time with milliseconds to datetime object.

    Args:
        epoch (int, optional): Epoch time in milliseconds. If None, returns None.

    Returns:
        datetime, optional: Corresponding datetime object, or None if input is None

    Example:
        >>> epoch_time = 1609459200000  # 2021-01-01 00:00:00 UTC
        >>> dt_obj = convert_epoch_millisecond_to_datetime(epoch_time)
        >>> print(dt_obj)  # 2021-01-01 00:00:00
    """
    return dt.datetime.fromtimestamp(epoch / 1000.0) if epoch else None


def convert_datetime_to_epoch_millisecond(
    datetime: dt.datetime | None,
) -> int | None:
    """
    Convert datetime object to Epoch time with milliseconds.

    Args:
        datetime (datetime, optional): Datetime object to convert. If None, returns None.

    Returns:
        int, optional: Epoch time in milliseconds, or None if input is None

    Example:
        >>> import datetime as dt
        >>> dt_obj = dt.datetime(2021, 1, 1)
        >>> epoch = convert_datetime_to_epoch_millisecond(dt_obj)
        >>> print(epoch)  # 1609459200000
    """
    return int(datetime.timestamp() * 1000) if datetime else None


def convert_string_to_datetime(datestr: str | None) -> dt.datetime | None:
    """
    Convert a date string to datetime object using flexible parsing.

    Args:
        datestr (str, optional): Date string to parse. If None or empty, returns None.

    Returns:
        datetime, optional: Parsed datetime object, or None if input is None/empty

    Raises:
        ImportError: If dateutil is not available
        ValueError: If the date string cannot be parsed

    Example:
        >>> dt_obj = convert_string_to_datetime("2021-01-01 12:30:00")
        >>> dt_obj = convert_string_to_datetime("Jan 1, 2021")
        >>> dt_obj = convert_string_to_datetime("2021/01/01")
    """
    if not datestr:
        return None

    return date_parser.parse(datestr)


def convert_python_to_ast_module(
    python_str: str | None = None,
    python_file_path: str | None = None,
    return_str: bool = False,
) -> ast.Module | str:
    """
    Parse Python code string or file and return its AST module.

    Args:
        python_str (str, optional): Python code string to parse
        python_file_path (str, optional): Path to Python file to parse
        return_str (bool): If True, return the source string instead of AST

    Returns:
        ast.Module or str: AST module of the parsed code, or source string if return_str=True

    Raises:
        ValueError: If neither python_str nor python_file_path is provided
        FileNotFoundError: If python_file_path doesn't exist
        SyntaxError: If the Python code has syntax errors

    Example:
        >>> code = "def hello(): return 'world'"
        >>> ast_module = convert_python_to_ast_module(python_str=code)
        >>> functions = extract_ast_functions(ast_module)
    """
    if not python_str and python_file_path:
        try:
            with open(python_file_path, encoding="utf-8") as source:
                python_str = source.read()
        except FileNotFoundError:
            raise FileNotFoundError(f"Python file not found: {python_file_path}")

    if not python_str:
        raise ValueError("Must provide either python_str or python_file_path")

    if return_str:
        return python_str

    return ast.parse(python_str)


def extract_ast_functions(ast_module: ast.Module) -> list[ast.FunctionDef]:
    """
    Extract all function definitions from an AST module.

    Args:
        ast_module (ast.Module): AST module to extract functions from

    Returns:
        list[ast.FunctionDef]: list of function definition nodes

    Example:
        >>> code = '''
        ... def func1():
        ...     pass
        ...
        ... class MyClass:
        ...     def method1(self):
        ...         pass
        ...
        ... def func2():
        ...     pass
        ... '''
        >>> ast_module = convert_python_to_ast_module(python_str=code)
        >>> functions = extract_ast_functions(ast_module)
        >>> print(len(functions))  # 2 (only top-level functions)

    Note:
        This only extracts top-level function definitions, not methods
        inside classes.
    """
    return [node for node in ast_module.body if isinstance(node, ast.FunctionDef)]


def convert_programming_text_to_title_case(clean_str: str) -> str:
    """
    Convert function names from programming conventions to human-readable display format.

    Transforms snake_case and camelCase function names into Title Case format suitable
    for user interfaces. Preserves leading underscores as spaces to maintain private
    function indicators.

    Args:
        clean_str (str): The original function name in snake_case or camelCase

    Returns:
        str: The formatted display name in Title Case

    Examples:
        >>> convert_programming_text_to_title_case('getUserData')
        'Get User Data'
        >>> convert_programming_text_to_title_case('calculate_total_sum')
        'Calculate Total Sum'
        >>> convert_programming_text_to_title_case('_private_method')
        ' Private Method'

    Note:
        Leading underscores are converted to spaces to preserve the indication
        that these are private/internal methods.
    """
    leading_underscores = ""
    working_str = clean_str

    # Extract leading underscores and convert to spaces
    while working_str.startswith("_"):
        leading_underscores += " "
        working_str = working_str[1:]

    # Convert camelCase to spaced format
    spaced_name = re.sub(r"([a-z])([A-Z])", r"\1 \2", working_str)
    # Convert snake_case to spaced format
    spaced_name = spaced_name.replace("_", " ")

    # Capitalize each word and join
    return leading_underscores + " ".join(
        word.capitalize() for word in spaced_name.split()
    )


def convert_snake_to_pascal(clean_str: str) -> str:
    """
    Convert snake_case string to camelCase (pascal case with lowercase first letter).

    Args:
        clean_str (str): Snake case string to convert

    Returns:
        str: String in camelCase format

    Example:
        >>> convert_snake_to_pascal('user_name_field')
        'userNameField'
        >>> convert_snake_to_pascal('api_key')
        'apiKey'
    """
    clean_str = clean_str.replace("_", " ").title().replace(" ", "")
    return clean_str[0].lower() + clean_str[1:] if clean_str else ""


def convert_str_to_snake_case(
    text_str: str, is_only_alphanumeric: bool = False, is_pascal: bool = False
) -> str:
    """
    Convert various string formats to snake_case.

    Can handle conversion from PascalCase/camelCase to snake_case, and optionally
    filter to only alphanumeric characters.

    Args:
        text_str (str): String to convert to snake_case
        is_only_alphanumeric (bool): If True, remove non-alphanumeric characters
        is_pascal (bool): If True, treat input as PascalCase/camelCase

    Returns:
        str: String converted to snake_case format

    Example:
        >>> convert_str_to_snake_case('UserNameField', is_pascal=True)
        'user_name_field'
        >>> convert_str_to_snake_case('User Name Field')
        'user_name_field'
        >>> convert_str_to_snake_case('User-Name!Field', is_only_alphanumeric=True)
        'usernamefield'
    """
    if is_pascal:
        # Convert PascalCase/camelCase to snake_case
        text_str = re.sub(r"(?<!^)(?=[A-Z])", "_", text_str)

    # Replace spaces with underscores and convert to lowercase
    text_str = text_str.replace(" ", "_").lower()

    if is_only_alphanumeric:
        # Remove all non-alphanumeric characters
        text_str = re.sub(r"\W+", "", text_str)

    return text_str


def is_valid_email(email: str) -> bool:
    """
    Test if provided string is a valid email format.

    Uses regex pattern matching to validate email format according to
    standard email format requirements.

    Args:
        email (str): Email string to validate

    Returns:
        bool: True if email format is valid

    Raises:
        InvalidEmail: If email format is invalid

    Example:
        >>> is_valid_email("user@example.com")
        True
        >>> is_valid_email("invalid-email")
        InvalidEmail: Invalid email format: "invalid-email"
    """
    pattern = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b"

    if re.fullmatch(pattern, email):
        return True
    else:
        raise InvalidEmailError(email=email)


def convert_string_to_bool(v: str | bool) -> bool:
    """
    Convert string representation to boolean value.

    Recognizes common string representations of boolean values and converts
    them to actual boolean type.

    Args:
        v (str or bool): Value to convert to boolean

    Returns:
        bool: Converted boolean value

    Example:
        >>> convert_string_to_bool("yes")
        True
        >>> convert_string_to_bool("false")
        False
        >>> convert_string_to_bool("1")
        True
        >>> convert_string_to_bool(True)
        True
    """
    if isinstance(v, bool):
        return v
    return str(v).lower() in ("yes", "true", "t", "1")


def concat_list_dataframe(df_ls: list[object]) -> object:
    """
    Take a list of DataFrames and concatenate them into one DataFrame.

    Args:
        df_ls (list[Any]): list of pandas DataFrames to concatenate

    Returns:
        Any: Concatenated DataFrame (returns Any due to optional pandas dependency)

    Raises:
        ImportError: If pandas is not available
        ConcatDataframe_InvalidElement: If any element is not a DataFrame

    Example:
        >>> df1 = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})
        >>> df2 = pd.DataFrame({'A': [5, 6], 'B': [7, 8]})
        >>> result = concat_list_dataframe([df1, df2])
        >>> print(len(result))  # 4 rows

    Note:
        Requires pandas to be installed. Uses inner join for concatenation
        and resets the index.
    """
    df = None
    for elem in df_ls:
        if not isinstance(elem, pd.DataFrame):
            raise ConcatDataframeError(elem)

        if len(elem.index) == 0:
            continue

        if df is None:
            df = elem
        else:
            df = pd.concat([df, elem], join="inner").reset_index(drop=True)

    return df


def merge_dict(
    source: dict[str, object], destination: dict[str, object]
) -> dict[str, object]:
    """
    Deep merge source dictionary into destination dictionary.

    Recursively merges nested dictionaries, with source values taking precedence
    over destination values for conflicts.

    Args:
        source (dict[str, Any]): Dictionary to merge from
        destination (dict[str, Any]): Dictionary to merge into (modified in place)

    Returns:
        dict[str, Any]: The merged destination dictionary

    Example:
        >>> dest = {"a": 1, "b": {"c": 2, "d": 3}}
        >>> src = {"b": {"d": 4, "e": 5}, "f": 6}
        >>> result = merge_dict(src, dest)
        >>> print(result)
        # {"a": 1, "b": {"c": 2, "d": 4, "e": 5}, "f": 6}

    Note:
        The destination dictionary is modified in place. For nested dictionaries,
        the merge is recursive.
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # Get existing node or create empty dict
            node = destination.setdefault(key, {})
            merge_dict(value, node)
        else:
            destination[key] = value

    return destination
